# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A mock IQ backend for testing."""
import datetime
from abc import abstractmethod
from typing import Sequence, List, Tuple, Dict, Union, Any

import numpy as np

from qiskit import QuantumCircuit
from qiskit.circuit.library import XGate, SXGate
from qiskit.result import Result
from qiskit.providers import BackendV2, Provider, convert_to_target
from qiskit.providers.fake_provider import FakeOpenPulse2Q
from qiskit.qobj.utils import MeasLevel

from qiskit_experiments.exceptions import QiskitError
from qiskit_experiments.framework import Options
from qiskit_experiments.test.utils import FakeJob
from qiskit_experiments.data_processing.exceptions import DataProcessorError
from qiskit_experiments.test.mock_iq_helpers import (
    MockIQExperimentHelper,
    MockIQParallelExperimentHelper,
    IQPoint,
)


class FakeOpenPulse2QV2(BackendV2):
    """BackendV2 conversion of qiskit.providers.fake_provider.FakeOpenPulse2Q"""

    def __init__(
        self,
        provider: Provider = None,
        name: str = None,
        description: str = None,
        online_date: datetime.datetime = None,
        backend_version: str = None,
        **fields,
    ):
        # Temporary workaround for missing support in Qiskit and qiskit-ibm-runtime
        from qiskit_experiments.test.patching import patch_sampler_test_support

        patch_sampler_test_support()

        super().__init__(provider, name, description, online_date, backend_version, **fields)

        backend_v1 = FakeOpenPulse2Q()
        # convert_to_target requires the description attribute
        backend_v1._configuration.description = "A fake test backend with pulse defaults"

        self._target = convert_to_target(
            backend_v1.configuration(),
            backend_v1.properties(),
            backend_v1.defaults(),
            add_delay=True,
        )
        # See commented out defaults() method below
        self._defaults = backend_v1._defaults

    # This method is not defined in the base class as we would like to avoid
    # relying on it as much as necessary. Individual tests should add it when
    # necessary.
    # def defaults(self):
    #     """Pulse defaults"""
    #     return self._defaults

    @property
    def max_circuits(self):
        return 300

    @property
    def target(self):
        return self._target


class MockRestlessBackend(FakeOpenPulse2QV2):
    """An abstract backend for testing that can mock restless data."""

    def __init__(self, rng_seed: int = 0):
        """
        Initialize the backend.
        """
        self._rng = np.random.default_rng(rng_seed)
        self._precomputed_probabilities = None
        super().__init__()

    @classmethod
    def _default_options(cls):
        """Default options of the test backend."""
        return Options(
            shots=1024,
            meas_level=MeasLevel.CLASSIFIED,
            meas_return="single",
        )

    @staticmethod
    def _get_state_strings(n_qubits: int) -> List[str]:
        """Generate all state strings for the system."""
        format_str = "{0:0" + str(n_qubits) + "b}"
        return list(format_str.format(state_num) for state_num in range(2**n_qubits))

    @abstractmethod
    def _compute_outcome_probabilities(self, circuits: List[QuantumCircuit]):
        """Compute the probabilities of measuring 0 or 1 for each of the given
         circuits based on the previous measurement shot.

        This methods computes the dictionary self._precomputed_probabilities where
        the keys are a tuple consisting of the circuit index and the previous outcome,
        e.g. "0" or "1" for a single qubit. The values are the corresponding probabilities.

        Args:
            circuits: The circuits from which to compute the probabilities.
        """

    def run(self, run_input, **options):
        """Run the restless backend."""

        self.options.update_options(**options)
        shots = self.options.get("shots")
        meas_level = self.options.get("meas_level")

        result = {
            "backend_name": f"{self.__class__.__name__}",
            "backend_version": "0",
            "qobj_id": 0,
            "job_id": 0,
            "success": True,
            "results": [],
        }

        self._compute_outcome_probabilities(run_input)

        if run_input[0].num_qubits != 2:
            raise DataProcessorError(f"{self.__class__.__name__} is a two qubit mock device.")

        prev_outcome, state_strings = "00", self._get_state_strings(2)

        # Setup the list of dicts where each dict corresponds to a circuit.
        sorted_memory = [{"memory": [], "metadata": circ.metadata} for circ in run_input]

        for _ in range(shots):
            for circ_idx, _ in enumerate(run_input):
                probs = self._precomputed_probabilities[(circ_idx, prev_outcome)]
                # Generate the next shot dependent on the pre-computed probabilities.
                outcome = self._rng.choice(state_strings, p=probs)
                # Append the single shot to the memory of the corresponding circuit.
                sorted_memory[circ_idx]["memory"].append(hex(int(outcome, 2)))

                prev_outcome = outcome

        for idx, circ in enumerate(run_input):
            counts = {}
            for key1, key2 in zip(["00", "01", "10", "11"], ["0x0", "0x1", "0x2", "0x3"]):
                counts[key1] = sorted_memory[idx]["memory"].count(key2)
            run_result = {
                "shots": shots,
                "success": True,
                "header": {"metadata": circ.metadata},
                "meas_level": meas_level,
                "data": {
                    "counts": counts,
                    "memory": sorted_memory[idx]["memory"],
                },
            }

            result["results"].append(run_result)

        return FakeJob(self, Result.from_dict(result))


class MockRestlessFineAmp(MockRestlessBackend):
    """A mock backend for restless single-qubit fine amplitude experiments."""

    def __init__(
        self, angle_error: float, angle_per_gate: float, gate_name: str, rng_seed: int = 0
    ):
        """Setup a mock backend to test the restless fine amplitude calibration.

        Args:
            angle_error: The rotation error per gate.
            angle_per_gate: The angle per gate.
            gate_name: The name of the gate to find in the circuit.
            rng_seed: The random bit generator seed.
        """
        self.angle_error = angle_error
        self._gate_name = gate_name
        self._angle_per_gate = angle_per_gate
        super().__init__(rng_seed=rng_seed)

        self.target.add_instruction(SXGate(), properties={(0,): None})
        self.target.add_instruction(XGate(), properties={(0,): None})

    def _compute_outcome_probabilities(self, circuits: List[QuantumCircuit]):
        """Compute the probabilities of being in the excited state or
        ground state for all circuits."""

        self._precomputed_probabilities = {}

        for idx, circuit in enumerate(circuits):

            n_ops = circuit.count_ops().get(self._gate_name, 0)
            angle = n_ops * (self._angle_per_gate + self.angle_error)

            if self._gate_name != "sx":
                angle += np.pi / 2 * circuit.count_ops().get("sx", 0)

            if self._gate_name != "x":
                angle += np.pi * circuit.count_ops().get("x", 0)

            prob_1 = np.sin(angle / 2) ** 2
            prob_0 = 1 - prob_1

            self._precomputed_probabilities[(idx, "00")] = [prob_0, prob_1, 0, 0]
            self._precomputed_probabilities[(idx, "01")] = [prob_1, prob_0, 0, 0]


class MockIQBackend(FakeOpenPulse2QV2):
    """A mock backend for testing with IQ data."""

    def __init__(
        self,
        experiment_helper: MockIQExperimentHelper = None,
        rng_seed: int = 0,
    ):
        """
        Initialize the backend.

        Args:
            experiment_helper(MockIQExperimentHelper): Experiment helper class that contains
                :meth:`~MockIQExperimentHelper.compute_probabilities` and
                :meth:`~MockIQExperimentHelper.iq_phase` methods for the backend to execute.
            rng_seed(int): The random seed value.
        """

        self._experiment_helper = experiment_helper
        self._rng = np.random.default_rng(rng_seed)
        self.simulator = True

        super().__init__()

    @classmethod
    def _default_options(cls):
        """Default options of the test backend."""
        return Options(
            shots=1024,
            meas_level=MeasLevel.KERNELED,
            meas_return="single",
        )

    @property
    def experiment_helper(self):
        """return the 'experiment_helper' attribute"""
        return self._experiment_helper

    @experiment_helper.setter
    def experiment_helper(self, value):
        """
        Setter for the experiment helper.
        Args:
            value(MockIQExperimentHelper): The helper for the backend to use for generating IQ shots.

        Raises:
            ValueError: Raised if the value to set is not of type `MockIQExperimentHelper`
        """
        cls = MockIQExperimentHelper
        if not isinstance(value, cls):
            raise ValueError(
                f"The input type is {str(type(value))} while the type expected type is "
                f"<{str(type(cls()))}>."
            )
        self._experiment_helper = value

    @staticmethod
    def _verify_parameters(output_length: int, prob_dict: Dict[str, float]):
        if output_length < 1:
            raise ValueError(f"The output length {output_length} is smaller than 1.")

        if not np.allclose(1, sum(prob_dict.values())):
            raise ValueError("The probabilities given don't sum up to 1.")
        for key in prob_dict.keys():
            if output_length is not len(key):
                raise ValueError(
                    "The output lengths of the circuit and the output lengths in the dictionary"
                    " don't match."
                )

    def _get_normal_samples_for_shot(
        self,
        qubits: Sequence[int],
    ) -> np.ndarray:
        """
        Produce a list in the size of num_qubits. Each entry value is produced from normal distribution
        with expected value of '0' and standard deviation of 1. The intention is that these samples are
        scaled by :func:`_scale_samples_for_widths` for various circuits, experiments, and their IQ
        widths; removing the need to query a RNG for each new width list.

        Example:
            .. code-block::
                # Generate template data
                template_iq_data = [np.nan] * shots
                for i_shot in range(n_shots):
                    real_data = self._get_normal_samples_for_shot(qubits)
                    imag_data = self._get_normal_samples_for_shot(qubits)
                    template_iq_data[i_shot] = np.array([real_data, imag_data], dtype="float").T

                # Scale template data to separate widths
                iq_data_1 = self._scale_samples_for_widths(template_iq_data, widths_1)
                iq_data_2 = self._scale_samples_for_widths(template_iq_data, widths_2)

                # IQ data should then be indexed randomly so that repeated usage does not give the same
                # order of samples.
                iq_data_circuit_1 = iq_data_1[random_indices_1]
                iq_data_circuit_2a = iq_data_2[random_indices_2a]
                iq_data_circuit_2b = iq_data_2[random_indices_2b]

        Args:
            num_qubits: The number of qubits in the circuit.

        Returns:
            Ndarray: A numpy array with values that were produced from normal distribution.
        """
        samples = [self._rng.normal(0, 1, size=1) for qubit in qubits]
        # we squeeze the second dimension because samples is List[qubit_number][0][0\1] = I\Q
        # and we want to change it to be List[qubit_number][0\1]
        return np.squeeze(np.array(samples), axis=1)

    def _scale_samples_for_widths(
        self, samples: List[np.ndarray], widths: List[float]
    ) -> List[np.ndarray]:
        """Scales `samples` by `widths` so that the data has the necessary std-dev.

        `samples` contains `n_shots` elements, each being :math:`n\times{}2` float values, representing
        the I and Q values for :math:`n` qubits. `widths` is a list of :math:`n` standard-deviations for
        each qubit. The IQ values for each list element in `samples` is scaled by the values in `widths`,
        for their respective qubits. It is assumed that the standard deviation of `samples` is :math:`1`.

        Args:
            samples: List of np.ndarrays containing random IQ samples for n qubits.
            widths: List of widths/standard-deviations to scale the data by.

        Returns:
            List: A list of samples with standard-deviations matching `widths`.
        """
        return [circ_samples * np.tile(widths, (2, 1)).T for circ_samples in samples]

    def _probability_dict_to_probability_array(
        self, prob_dict: Dict[str, float], num_qubits: int
    ) -> List[float]:
        prob_list = [0] * (2**num_qubits)
        for output_str, probability in prob_dict.items():
            index = int(output_str, 2)
            prob_list[index] = probability
        return prob_list

    def _draw_iq_shots(
        self,
        prob: List[float],
        shots: int,
        circ_qubits: Sequence[int],
        iq_cluster_centers: List[Tuple[IQPoint, IQPoint]],
        iq_cluster_width: List[float],
        phase: float = 0.0,
    ) -> List[List[List[Union[float, complex]]]]:
        """
        Produce an IQ shot.

        Args:
            prob: A list of probabilities for each output.
            shots: The number of times the circuit will run.
            circ_qubits: The qubits of the circuit.
            iq_cluster_centers: A list of tuples containing the clusters' centers in the IQ plane. There
            are different centers for different logical values of the qubit.
            iq_cluster_width: A list of standard deviation values for the sampling of each qubit.
            phase: The added phase needed to apply to the shot data.
        Returns:
            List[List[Tuple[float, float]]]: A list of shots. Each shot consists of a list of qubits.
            The qubits are tuples with two values [I,Q].
            The output structure is  - List[shot index][qubit index] = [I,Q]
        """
        # Randomize samples (width=1)
        qubits_iq_template_rand = [np.nan] * shots
        for shot in range(shots):
            rand_i = self._get_normal_samples_for_shot(circ_qubits)
            rand_q = self._get_normal_samples_for_shot(circ_qubits)
            qubits_iq_template_rand[shot] = np.array([rand_i, rand_q], dtype="float").T

        # Scale samples to use iq_cluster_width.
        exp_widths = [iq_cluster_width[i_qubit] for i_qubit in circ_qubits]
        qubits_iq_rand = self._scale_samples_for_widths(qubits_iq_template_rand, exp_widths)

        memory = []
        shot_num = 0

        for output_number, number_of_occurrences in enumerate(
            self._rng.multinomial(shots, prob, size=1)[0]
        ):
            state_str = str(format(output_number, "b").zfill(len(circ_qubits)))
            for _ in range(number_of_occurrences):
                shot_memory = []
                # the iteration on the string variable state_str starts from the MSB. For readability,
                # we will reverse the string so the loop will run from the LSB to MSB.
                for iq_center, qubit_iq_rand_sample, char_qubit in zip(
                    iq_cluster_centers, qubits_iq_rand[shot_num], state_str[::-1]
                ):
                    # The structure of iq_cluster_centers is [qubit_number][logic_result][I/Q].
                    i_center = iq_center[int(char_qubit)][0]
                    q_center = iq_center[int(char_qubit)][1]

                    point_i = i_center + qubit_iq_rand_sample[0]
                    point_q = q_center + qubit_iq_rand_sample[1]

                    # Adding phase if not 0.0
                    if not np.allclose(phase, 0.0):
                        complex_iq = (point_i + 1.0j * point_q) * np.exp(1.0j * phase)
                        point_i, point_q = np.real(complex_iq), np.imag(complex_iq)

                    shot_memory.append([point_i, point_q])
                # We proceed to the next occurrence - meaning it's a new shot.
                memory.append(shot_memory)
                shot_num += 1

        return memory

    def _generate_data(
        self, prob_dict: Dict[str, float], circuit: QuantumCircuit
    ) -> Dict[str, Any]:
        """
        Generate data for the circuit.

        Args:
            prob_dict: A dictionary whose keys are strings representing the output vectors and
            their values are the probability to get the output in this circuit.
            circuit: The circuit that needs to be simulated.

        Returns:
            A dictionary that's filled with the simulated data. The output format is different between
            measurement level 1 and measurement level 2.
        """
        # The output is proportional to the number of classical bit.
        output_length = int(np.sum([creg.size for creg in circuit.cregs]))
        self._verify_parameters(output_length, prob_dict)
        prob_arr = self._probability_dict_to_probability_array(prob_dict, output_length)
        shots = self.options.get("shots")
        meas_level = self.options.get("meas_level")
        meas_return = self.options.get("meas_return")
        run_result = {}

        if meas_level == MeasLevel.CLASSIFIED:
            counts = {}
            results = self._rng.multinomial(shots, prob_arr, size=1)[0]
            for result, num_occurrences in enumerate(results):
                result_in_str = str(format(result, "b").zfill(output_length))
                counts[result_in_str] = num_occurrences
            run_result["counts"] = counts
            if meas_return == "single" or self.options.get("memory"):
                run_result["memory"] = [
                    format(result, "x") for result, num in enumerate(results) for _ in range(num)
                ]
        else:
            # Phase has meaning only for IQ shot, so we calculate it here
            phase = self.experiment_helper.iq_phase([circuit])[0]
            iq_cluster_centers, iq_cluster_width = self.experiment_helper.iq_clusters([circuit])[0]

            # 'circ_qubits' get a list of all the qubits
            memory = self._draw_iq_shots(
                prob_arr,
                shots,
                list(range(output_length)),
                iq_cluster_centers,
                iq_cluster_width,
                phase,
            )
            if meas_return == "avg":
                memory = np.average(np.array(memory), axis=0).tolist()

            run_result["memory"] = memory
        return run_result

    def run(self, run_input: List[QuantumCircuit], **run_options) -> FakeJob:
        """
        Run the IQ backend.

        Args:
            run_input: A list of QuantumCircuit for which the backend will generate
                data.
            **run_options: Experiment running options. The options that are supported
                in this backend are `meas_level`, `meas_return` and `shots`:

                * meas_level: To generate data in the IQ plane, `meas_level` should be
                  assigned 1 or ``MeasLevel.KERNELED``. If `meas_level` is 2 or
                  ``MeasLevel.CLASSIFIED``, the generated data will be in the form
                  of `counts`.
                * meas_return: This option will only take effect if `meas_level` =
                  ``MeasLevel.CLASSIFIED``. It can get either
                  ``MeasReturnType.AVERAGE`` or ``MeasReturnType.SINGLE``. For
                  ``MeasReturnType.SINGLE``, the data of each shot will be stored in
                  the result. For ``MeasReturnType.AVERAGE``, an average of all the
                  shots will be calculated and stored in the result.
                * shots: The number of times the circuit will run.

        Returns:
            FakeJob: A job that contains the simulated data.

        Raises:
            QiskitError: Raised if the user try to run the experiment without setting a helper.
        """

        if not self.experiment_helper:
            raise QiskitError("The backend `experiment_helper` attribute cannot be 'None'.")

        self.options.update_options(**run_options)
        shots = self.options.get("shots")
        meas_level = self.options.get("meas_level")

        result = {
            "backend_name": f"{self.__class__.__name__}",
            "backend_version": "0",
            "qobj_id": "0",
            "job_id": "0",
            "success": True,
            "results": [],
        }
        prob_list = self.experiment_helper.compute_probabilities(run_input)
        for prob, circ in zip(prob_list, run_input):
            run_result = {
                "shots": shots,
                "success": True,
                "header": {"metadata": circ.metadata},
                "meas_level": meas_level,
            }

            run_result["data"] = self._generate_data(prob, circ)
            result["results"].append(run_result)

        return FakeJob(self, Result.from_dict(result))


class MockIQParallelBackend(MockIQBackend):
    """A mock backend for testing parallel experiments with IQ data."""

    def __init__(
        self,
        experiment_helper: MockIQParallelExperimentHelper = None,
        rng_seed: int = 0,
    ):
        """
        Initialize the backend.

        Args:
            experiment_helper: Parallel experiment helper class that contains
                helper classes for each experiment.
            rng_seed: The random seed value.
        """
        super().__init__(experiment_helper, rng_seed)

    @property
    def experiment_helper(self):
        """return the 'experiment_helper' attribute"""
        return self._experiment_helper

    @experiment_helper.setter
    def experiment_helper(self, value):
        """
        Setter for the experiment helper.

        Args:
            value(MockIQParallelExperimentHelper): The helper for the backend to use for generating IQ
             shots.

        Raises:
            ValueError: Raised if the value to set is not of type `MockIQExperimentHelper`
        """
        cls = MockIQParallelExperimentHelper
        if not isinstance(value, cls):
            raise ValueError(
                f"The input type is {str(type(value))} while the type expected type is <{str(cls)}>."
            )
        self._experiment_helper = value

    def _parallel_draw_iq_shots(
        self,
        list_exp_dict: List[Dict[str, Union[List, int]]],
        shots: int,
        circ_qubits: List[int],
        circ_idx: int,
    ) -> List[List[List[Union[float, complex]]]]:
        """
        Produce an IQ shot.
        Args:
            list_exp_dict: A list of dictionaries for each experiment. It is determined by the
                ``MockIQParallelExperimentHelper`` object provided to the backend.
            shots: The number of times the circuit will run.
            circ_qubits: List of qubits that are used in this circuit.
            circ_idx: The circuit index.

        Returns:
            List[List[Tuple[float, float]]]: A list of shots. Each shot consists of a list of qubits.
            The qubits are tuples with two values [I,Q].
            The output structure is  - List[shot index][qubit index] = [I,Q]
        """
        # Randomize samples (width=1)
        qubits_iq_template_rand = [np.nan] * shots
        for shot in range(shots):
            rand_i = self._get_normal_samples_for_shot(circ_qubits)
            rand_q = self._get_normal_samples_for_shot(circ_qubits)
            qubits_iq_template_rand[shot] = np.array([rand_i, rand_q], dtype="float").T

        memory = [[] for _ in range(shots)]

        # The use of idx_shift is to sample 'qubits_iq_rand' correctly
        sample_idx_shift = 0

        # The code generates data as follows:
        # for each experiment, it first checks if it needs to generate data for it. If it does, then the
        # multinomial probability function draws lots for all the shots, and we store this data in the
        # corresponding position in the output list. After that we move on to the next experiment.
        for exp_dict in list_exp_dict:
            # skipping experiments that don't need data generation for this circuit.
            if exp_dict["num_circuits"] <= circ_idx:
                continue

            qubits = list(exp_dict["physical_qubits"])
            prob = self._probability_dict_to_probability_array(
                exp_dict["prob"][circ_idx], len(qubits)
            )
            phase = exp_dict["phase"][circ_idx]
            iq_centers = exp_dict["centers"][circ_idx]
            iq_widths = exp_dict["widths"][circ_idx]
            exp_widths = [iq_widths[i_qubit] for i_qubit in circ_qubits]

            # Rescale samples to appropriate width for the given parallel circuits
            qubits_iq_rand = self._scale_samples_for_widths(qubits_iq_template_rand, exp_widths)

            shot_num = 0

            for output_number, number_of_occurrences in enumerate(
                self._rng.multinomial(shots, prob, size=1)[0]
            ):
                state_str = str(format(output_number, "b").zfill(len(qubits)))
                for _ in range(number_of_occurrences):
                    # the iteration on the string variable state_str starts from the MSB. For
                    # readability, we will reverse the string so the loop will run from the LSB to MSB.
                    for qubit_idx, qubit, char_qubit in zip(
                        range(len(qubits)), qubits, state_str[::-1]
                    ):

                        i_center = iq_centers[qubit][int(char_qubit)][0]
                        q_center = iq_centers[qubit][int(char_qubit)][1]

                        # we use 'sample_idx_shift' to take the sample corresponding to the current qubit
                        # in 'qubits_iq_rand[shot_num]'.
                        point_i = (
                            i_center + qubits_iq_rand[shot_num][qubit_idx + sample_idx_shift][0]
                        )
                        point_q = (
                            q_center + qubits_iq_rand[shot_num][qubit_idx + sample_idx_shift][1]
                        )

                        # Adding phase if not 0.0
                        if not np.allclose(phase, 0.0):
                            complex_iq = (point_i + 1.0j * point_q) * np.exp(1.0j * phase)
                            point_i, point_q = np.real(complex_iq), np.imag(complex_iq)

                        memory[shot_num].append([point_i, point_q])
                    shot_num += 1

            sample_idx_shift = sample_idx_shift + len(qubits)
        return memory

    def _parallel_generate_data(
        self,
        list_exp_dict: List[Dict[str, Union[List, int]]],
        circ_idx: int,
    ) -> Dict[str, Any]:
        """
        Generate data for the circuit.
        Args:
            list_exp_dict (List): A List of dictionaries, each dictionary contains data of an experiment.
            circ_idx (int): The circuit number we simulate.

        Returns:
            A dictionary that's filled with the simulated data.

        Raises:
            QiskitError: Raising an error if in the experiment running option, classified data is
            requested.
        """
        circ_qubit_list = []
        for exp_dict in list_exp_dict:
            if circ_idx < exp_dict["num_circuits"]:
                circ_qubit_list = circ_qubit_list + list(exp_dict["physical_qubits"])

        shots = self.options.get("shots")
        meas_level = self.options.get("meas_level")
        meas_return = self.options.get("meas_return")
        run_result = {}

        if meas_level == MeasLevel.KERNELED:
            memory = self._parallel_draw_iq_shots(list_exp_dict, shots, circ_qubit_list, circ_idx)
            if meas_return == "avg":
                memory = np.average(np.array(memory), axis=0).tolist()

            run_result["memory"] = memory
        else:
            # The backend doesn't currently support 'meas_level = MeasLevel.CLASSIFIED'.
            raise QiskitError("Classified data generator isn't supported for this backend")

        return run_result

    def run(self, run_input: List[QuantumCircuit], **run_options) -> FakeJob:
        """
        Run the IQ backend.

        Args:
            run_input: A list of QuantumCircuit for which the backend will generate
                data.
            **run_options: Experiment running options. The options that are supported
                in this backend are `meas_level`, `meas_return` and `shots`:

                * meas_level: To generate data in the IQ plane, `meas_level` should be
                  assigned 1 or ``MeasLevel.KERNELED``. If `meas_level` is 2 or
                  ``MeasLevel.CLASSIFIED``, the generated data will be in the form
                  of `counts`.
                * meas_return: This option will only take effect if `meas_level` =
                  ``MeasLevel.CLASSIFIED``. It can get either
                  ``MeasReturnType.AVERAGE`` or ``MeasReturnType.SINGLE``. For
                  ``MeasReturnType.SINGLE``, the data of each shot will be stored in
                  the result. For ``MeasReturnType.AVERAGE``, an average of all the
                  shots will be calculated and stored in the result.
                * shots: The number of times the circuit will run.

        Returns:
            FakeJob: A job that contains the simulated data.

        Raises:
            QiskitError: Raised if the user try to run the experiment without setting a helper.
        """

        if not self.experiment_helper:
            raise QiskitError("The backend `experiment_helper` attribute cannot be 'None'.")

        self.options.update_options(**run_options)
        shots = self.options.get("shots")
        meas_level = self.options.get("meas_level")

        result = {
            "backend_name": f"{self.__class__.__name__}",
            "backend_version": "0",
            "qobj_id": "0",
            "job_id": "0",
            "success": True,
            "results": [],
        }

        experiment_data_list = self.experiment_helper.compute_probabilities(run_input)
        for circ_idx, circ in enumerate(run_input):
            run_result = {
                "shots": shots,
                "success": True,
                "header": {"metadata": circ.metadata},
                "meas_level": meas_level,
            }

            run_result["data"] = self._parallel_generate_data(experiment_data_list, circ_idx)
            result["results"].append(run_result)

        return FakeJob(self, Result.from_dict(result))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Base Experiment class.
"""

from abc import ABC, abstractmethod
import copy
from collections import OrderedDict
from typing import Sequence, Optional, Tuple, List, Dict, Union

from qiskit import transpile, QuantumCircuit
from qiskit.providers import Job, Backend
from qiskit.exceptions import QiskitError
from qiskit.qobj.utils import MeasLevel
from qiskit.providers.options import Options
from qiskit.primitives.base import BaseSamplerV2
from qiskit_ibm_runtime import SamplerV2 as Sampler
from qiskit_experiments.framework import BackendData
from qiskit_experiments.framework.store_init_args import StoreInitArgs
from qiskit_experiments.framework.base_analysis import BaseAnalysis
from qiskit_experiments.framework.experiment_data import ExperimentData
from qiskit_experiments.framework.configs import ExperimentConfig
from qiskit_experiments.database_service import Qubit


class BaseExperiment(ABC, StoreInitArgs):
    """Abstract base class for experiments."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        analysis: Optional[BaseAnalysis] = None,
        backend: Optional[Backend] = None,
        experiment_type: Optional[str] = None,
        backend_run: Options[bool] = False,
    ):
        """Initialize the experiment object.

        Args:
            physical_qubits: list of physical qubits for the experiment.
            analysis: Optional, the analysis to use for the experiment.
            backend: Optional, the backend to run the experiment on.
            experiment_type: Optional, the experiment type string.
            backend_run: Optional, use backend run vs the sampler (temporary)
        Raises:
            QiskitError: If qubits contains duplicates.
        """
        # Experiment identification metadata
        self.experiment_type = experiment_type

        # Circuit parameters
        self._num_qubits = len(physical_qubits)
        self._physical_qubits = tuple(physical_qubits)
        if self._num_qubits != len(set(self._physical_qubits)):
            raise QiskitError("Duplicate qubits in physical qubits list.")

        # Experiment options
        self._experiment_options = self._default_experiment_options()
        self._transpile_options = self._default_transpile_options()
        self._run_options = self._default_run_options()

        # Store keys of non-default options
        self._set_experiment_options = set()
        self._set_transpile_options = set()
        self._set_run_options = set()
        self._set_analysis_options = set()

        # Set analysis
        self._analysis = None
        if analysis:
            self.analysis = analysis

        # Set backend
        # This should be called last in case `_set_backend` access any of the
        # attributes created during initialization
        self._backend = None
        self._backend_data = None
        self._backend_run = backend_run
        if isinstance(backend, Backend):
            self._set_backend(backend)

    @property
    def experiment_type(self) -> str:
        """Return experiment type."""
        return self._type

    @experiment_type.setter
    def experiment_type(self, exp_type: str) -> None:
        """Set the type for the experiment."""
        if exp_type is None:
            self._type = type(self).__name__
        else:
            self._type = exp_type

    @property
    def physical_qubits(self) -> Tuple[int, ...]:
        """Return the device qubits for the experiment."""
        return self._physical_qubits

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits for the experiment."""
        return self._num_qubits

    @property
    def analysis(self) -> Union[BaseAnalysis, None]:
        """Return the analysis instance for the experiment"""
        return self._analysis

    @analysis.setter
    def analysis(self, analysis: Union[BaseAnalysis, None]) -> None:
        """Set the analysis instance for the experiment"""
        if analysis is not None and not isinstance(analysis, BaseAnalysis):
            raise TypeError("Input is not a None or a BaseAnalysis subclass.")
        self._analysis = analysis

    @property
    def backend(self) -> Union[Backend, None]:
        """Return the backend for the experiment"""
        return self._backend

    @backend.setter
    def backend(self, backend: Union[Backend, None]) -> None:
        """Set the backend for the experiment"""
        if not isinstance(backend, Backend):
            raise TypeError("Input is not a backend.")
        self._set_backend(backend)

    def _set_backend(self, backend: Backend):
        """Set the backend for the experiment.

        Subclasses can override this method to extract additional
        properties from the supplied backend if required.
        """
        self._backend = backend
        self._backend_data = BackendData(backend)

    def copy(self) -> "BaseExperiment":
        """Return a copy of the experiment"""
        # We want to avoid a deep copy be default for performance so we
        # need to also copy the Options structures so that if they are
        # updated on the copy they don't effect the original.
        ret = copy.copy(self)
        if self.analysis:
            ret.analysis = self.analysis.copy()

        ret._experiment_options = copy.copy(self._experiment_options)
        ret._run_options = copy.copy(self._run_options)
        ret._transpile_options = copy.copy(self._transpile_options)

        ret._set_experiment_options = copy.copy(self._set_experiment_options)
        ret._set_transpile_options = copy.copy(self._set_transpile_options)
        ret._set_run_options = copy.copy(self._set_run_options)
        return ret

    def config(self) -> ExperimentConfig:
        """Return the config dataclass for this experiment"""
        args = tuple(getattr(self, "__init_args__", OrderedDict()).values())
        kwargs = dict(getattr(self, "__init_kwargs__", OrderedDict()))
        # Only store non-default valued options
        experiment_options = dict(
            (key, getattr(self._experiment_options, key)) for key in self._set_experiment_options
        )
        transpile_options = dict(
            (key, getattr(self._transpile_options, key)) for key in self._set_transpile_options
        )
        run_options = dict((key, getattr(self._run_options, key)) for key in self._set_run_options)
        return ExperimentConfig(
            cls=type(self),
            args=args,
            kwargs=kwargs,
            experiment_options=experiment_options,
            transpile_options=transpile_options,
            run_options=run_options,
        )

    @classmethod
    def from_config(cls, config: Union[ExperimentConfig, Dict]) -> "BaseExperiment":
        """Initialize an experiment from experiment config"""
        if isinstance(config, dict):
            config = ExperimentConfig(**dict)
        ret = cls(*config.args, **config.kwargs)
        if config.experiment_options:
            ret.set_experiment_options(**config.experiment_options)
        if config.transpile_options:
            ret.set_transpile_options(**config.transpile_options)
        if config.run_options:
            ret.set_run_options(**config.run_options)
        return ret

    def run(
        self,
        backend: Optional[Backend] = None,
        sampler: Optional[BaseSamplerV2] = None,
        analysis: Optional[Union[BaseAnalysis, None]] = "default",
        timeout: Optional[float] = None,
        backend_run: Optional[bool] = None,
        **run_options,
    ) -> ExperimentData:
        """Run an experiment and perform analysis.

        Args:
            backend: Optional, the backend to run on. Will override existing backend settings.
            sampler: Optional, the sampler to run the experiment on.
                      If None then a sampler will be invoked from previously
                      set backend
            analysis: Optional, a custom analysis instance to use for performing
                      analysis. If None analysis will not be run. If ``"default"``
                      the experiments :meth:`analysis` instance will be used if
                      it contains one.
            timeout: Time to wait for experiment jobs to finish running before
                     cancelling.
            backend_run: Use backend run (temp option for testing)
            run_options: backend runtime options used for circuit execution.

        Returns:
            The experiment data object.

        Raises:
            QiskitError: If experiment is run with an incompatible existing
                         ExperimentData container.
        """

        if (
            (backend is not None)
            or (sampler is not None)
            or analysis != "default"
            or run_options
            or (backend_run is not None)
        ):
            # Make a copy to update analysis or backend if one is provided at runtime
            experiment = self.copy()
            if backend_run is not None:
                experiment._backend_run = backend_run
            # we specified a backend OR a sampler
            if (backend is not None) or (sampler is not None):
                if sampler is None:
                    # backend only specified
                    experiment._set_backend(backend)
                elif backend is None:
                    # sampler only specifid
                    experiment._set_backend(sampler._backend)
                else:
                    # we specified both a sampler and a backend
                    if self._backend_run:
                        experiment._set_backend(backend)
                    else:
                        experiment._set_backend(sampler._backend)
            if isinstance(analysis, BaseAnalysis):
                experiment.analysis = analysis
            if run_options:
                experiment.set_run_options(**run_options)
        else:
            experiment = self

        if experiment.backend is None:
            raise QiskitError("Cannot run experiment, no backend has been set.")

        # Finalize experiment before executions
        experiment._finalize()

        # Generate and transpile circuits
        transpiled_circuits = experiment._transpiled_circuits()

        # Initialize result container
        experiment_data = experiment._initialize_experiment_data()

        # Run options
        run_opts = experiment.run_options.__dict__

        # Run jobs
        jobs = experiment._run_jobs(transpiled_circuits, sampler=sampler, **run_opts)
        experiment_data.add_jobs(jobs, timeout=timeout)

        # Optionally run analysis
        if analysis and experiment.analysis:
            return experiment.analysis.run(experiment_data)
        else:
            return experiment_data

    def _initialize_experiment_data(self) -> ExperimentData:
        """Initialize the return data container for the experiment run"""
        return ExperimentData(experiment=self)

    def _finalize(self):
        """Finalize experiment object before running jobs.

        Subclasses can override this method to set any final option
        values derived from other options or attributes of the
        experiment before `_run` is called.
        """
        pass

    def _max_circuits(self, backend: Backend = None):
        """
        Calculate the maximum number of circuits per job for the experiment.
        """

        # set backend
        if backend is None:
            if self.backend is None:
                raise QiskitError("A backend must be provided.")
            backend = self.backend
        # Get max circuits for job splitting
        max_circuits_option = getattr(self.experiment_options, "max_circuits", None)
        max_circuits_backend = BackendData(backend).max_circuits

        if max_circuits_option and max_circuits_backend:
            return min(max_circuits_option, max_circuits_backend)
        elif max_circuits_option:
            return max_circuits_option
        else:
            return max_circuits_backend

    def job_info(self, backend: Backend = None):
        """
        Get information about job distribution for the experiment on a specific
        backend.

        Args:
            backend: Optional, the backend for which to get job distribution
                information. If not specified, the experiment must already have a
                set backend.

        Returns:
            dict: A dictionary containing information about job distribution.

                - "Total number of circuits in the experiment": Total number of
                  circuits in the experiment.

                - "Maximum number of circuits per job": Maximum number of
                  circuits in one job based on backend and experiment settings.

                - "Total number of jobs": Number of jobs needed to run this
                  experiment on the currently set backend.

        Raises:
            QiskitError: if backend is not specified.

        """
        max_circuits = self._max_circuits(backend)
        total_circuits = len(self.circuits())

        if max_circuits is None:
            num_jobs = 1
        else:
            num_jobs = (total_circuits + max_circuits - 1) // max_circuits
        return {
            "Total number of circuits in the experiment": total_circuits,
            "Maximum number of circuits per job": max_circuits,
            "Total number of jobs": num_jobs,
        }

    def _run_jobs(
        self, circuits: List[QuantumCircuit], sampler: BaseSamplerV2 = None, **run_options
    ) -> List[Job]:
        """Run circuits on backend as 1 or more jobs."""
        max_circuits = self._max_circuits(self.backend)

        # Run experiment jobs
        if max_circuits and (len(circuits) > max_circuits):
            # Split jobs for backends that have a maximum job size
            job_circuits = [
                circuits[i : i + max_circuits] for i in range(0, len(circuits), max_circuits)
            ]
        else:
            # Run as single job
            job_circuits = [circuits]

        # Run jobs
        if not self._backend_run:
            if sampler is None:
                # instantiate a sampler from the backend
                sampler = Sampler(self.backend)

                # have to hand set some of these options
                # see https://docs.quantum.ibm.com/api/qiskit-ibm-runtime
                # /qiskit_ibm_runtime.options.SamplerExecutionOptionsV2
                if "init_qubits" in run_options:
                    sampler.options.execution.init_qubits = run_options["init_qubits"]
                if "rep_delay" in run_options:
                    sampler.options.execution.rep_delay = run_options["rep_delay"]
                if "meas_level" in run_options:
                    if run_options["meas_level"] == 2:
                        sampler.options.execution.meas_type = "classified"
                    elif run_options["meas_level"] == 1:
                        if "meas_return" in run_options:
                            if run_options["meas_return"] == "avg":
                                sampler.options.execution.meas_type = "avg_kerneled"
                            else:
                                sampler.options.execution.meas_type = "kerneled"
                        else:
                            # assume this is what is wanted if no  meas return specified
                            sampler.options.execution.meas_type = "kerneled"
                    else:
                        raise QiskitError("Only meas level 1 + 2 supported by sampler")
                if "noise_model" in run_options:
                    sampler.options.simulator.noise_model = run_options["noise_model"]
                if "seed_simulator" in run_options:
                    sampler.options.simulator.seed_simulator = run_options["seed_simulator"]

                if run_options.get("shots") is not None:
                    sampler.options.default_shots = run_options.get("shots")

            jobs = [sampler.run(circs) for circs in job_circuits]
        else:
            jobs = [self.backend.run(circs, **run_options) for circs in job_circuits]

        return jobs

    @abstractmethod
    def circuits(self) -> List[QuantumCircuit]:
        """Return a list of experiment circuits.

        Returns:
            A list of :class:`~qiskit.circuit.QuantumCircuit`.

        .. note::
            These circuits should be on qubits ``[0, .., N-1]`` for an
            *N*-qubit experiment. The circuits mapped to physical qubits
            are obtained via the internal :meth:`_transpiled_circuits` method.
        """
        # NOTE: Subclasses should override this method using the `options`
        # values for any explicit experiment options that affect circuit
        # generation

    def _transpiled_circuits(self) -> List[QuantumCircuit]:
        """Return a list of experiment circuits, transpiled.

        This function can be overridden to define custom transpilation.
        """
        transpile_opts = copy.copy(self.transpile_options.__dict__)
        transpile_opts["initial_layout"] = list(self.physical_qubits)
        transpiled = transpile(self.circuits(), self.backend, **transpile_opts)

        return transpiled

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            max_circuits (Optional[int]): The maximum number of circuits per job when
                running an experiment on a backend.
        """
        # Experiment subclasses should override this method to return
        # an `Options` object containing all the supported options for
        # that experiment and their default values. Only options listed
        # here can be modified later by the different methods for
        # setting options.
        return Options(max_circuits=None)

    @property
    def experiment_options(self) -> Options:
        """Return the options for the experiment."""
        return self._experiment_options

    def set_experiment_options(self, **fields):
        """Set the experiment options.

        Args:
            fields: The fields to update the options

        Raises:
            AttributeError: If the field passed in is not a supported options
        """
        for field in fields:
            if not hasattr(self._experiment_options, field):
                raise AttributeError(
                    f"Options field {field} is not valid for {type(self).__name__}"
                )
        self._experiment_options.update_options(**fields)
        self._set_experiment_options = self._set_experiment_options.union(fields)

    @classmethod
    def _default_transpile_options(cls) -> Options:
        """Default transpiler options for transpilation of circuits"""
        # Experiment subclasses can override this method if they need
        # to set specific default transpiler options to transpile the
        # experiment circuits.
        return Options(optimization_level=0)

    @property
    def transpile_options(self) -> Options:
        """Return the transpiler options for the :meth:`run` method."""
        return self._transpile_options

    def set_transpile_options(self, **fields):
        """Set the transpiler options for :meth:`run` method.

        Args:
            fields: The fields to update the options

        Raises:
            QiskitError: If `initial_layout` is one of the fields.

        .. seealso:: The :ref:`guide_setting_options` guide for code example.
        """
        if "initial_layout" in fields:
            raise QiskitError(
                "Initial layout cannot be specified as a transpile option"
                " as it is determined by the experiment physical qubits."
            )
        self._transpile_options.update_options(**fields)
        self._set_transpile_options = self._set_transpile_options.union(fields)

    @classmethod
    def _default_run_options(cls) -> Options:
        """Default options values for the experiment :meth:`run` method."""
        return Options(meas_level=MeasLevel.CLASSIFIED)

    @property
    def run_options(self) -> Options:
        """Return options values for the experiment :meth:`run` method."""
        return self._run_options

    def set_run_options(self, **fields):
        """Set options values for the experiment  :meth:`run` method.

        Args:
            fields: The fields to update the options

        .. seealso:: The :ref:`guide_setting_options` guide for code example.
        """
        self._run_options.update_options(**fields)
        self._set_run_options = self._set_run_options.union(fields)

    def _metadata(self) -> Dict[str, any]:
        """Return experiment metadata for ExperimentData.

        By default, this assumes the experiment is running on qubits only. Subclasses can override
        this method to add custom experiment metadata to the returned experiment result data.
        """
        metadata = {
            "physical_qubits": list(self.physical_qubits),
            "device_components": list(map(Qubit, self.physical_qubits)),
        }
        return metadata

    def __json_encode__(self):
        """Convert to format that can be JSON serialized"""
        return self.config()

    @classmethod
    def __json_decode__(cls, value):
        """Load from JSON compatible format"""
        return cls.from_config(value)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Utilities for visualization."""
from typing import List, Optional, Tuple, Union

import numpy as np
from qiskit.exceptions import QiskitError

ExtentTuple = Tuple[float, float, float, float]


class DataExtentCalculator:
    """A class to handle computing the extent of two-dimensional data.

    This class makes computing the extent of two-dimensional data easier, especially for
    adding images to plots where the extent of the axes (i.e., axis limits) are not
    known prior to figure generation. An instance of this class can be used to compute
    an extent tuple which is larger than the registered data, and thus has a margin that
    can exceed the axis-limits or be used as the axis limits without cropping additional
    data-points from a figure. An extent tuple ``(x_min, x_max, y_min, y_max)`` contains
    the minimum and maximum values for the X and Y dimensions.

    Data is registered with a :class:`DataExtentCalculator` so that the computed extent
    covers all values in the data array. The extent tuple is computed as follows:

    1. The maximum and minimum values for input data is stored whenever new data
       arrays are registered. This is the data-extent: the minimum-area bounding box
       that contains all registered data.
    2. The data-extent is enlarged/shrunk by scaling its width and height by
       :attr:`multiplier`.
    3. If :attr:`aspect_ratio` is not ``None``, the scaled extent tuple is extended
       in one of the dimensions so that the output extent tuple is larger and the
       target aspect ratio is achieved.
    """

    def __init__(self, multiplier: float = 1.0, aspect_ratio: Optional[float] = 1.0):
        """Create an extent calculator.

        Args:
            multiplier: The factor by which to scale the extent of the data when
                computing the output extent tuple. Defaults to 1.0.
            aspect_ratio: An optional target aspect ratio for the output extent tuple.
                If None, the extent tuple is not extended to achieve a given aspect
                ratio. Defaults to None.
        """
        self._multiplier = multiplier
        self._aspect_ratio = aspect_ratio
        self._data_extent = np.asarray([[np.inf, -np.inf], [np.inf, -np.inf]])

    @property
    def multiplier(self) -> float:
        """The multiplier by which to scale the data-extent.

        Returns:
            The multiplier for the computed extent.
        """
        return self._multiplier

    @property
    def aspect_ratio(self) -> Optional[float]:
        """The target aspect ratio.

        If None, the :class:`DataExtentCalculator` instance will not modify the computed
        extent tuple to achieve a given aspect ratio; instead only scale by
        :attr:`multiplier`.

        Returns:
            The target aspect ratio of the computed extent tuple.
        """
        return self._aspect_ratio

    def register_data(self, data: Union[List, np.ndarray], dim: Optional[int] = None):
        r"""Register data to modify the resulting extent tuple.

        Args:
            data: Array or list of data values to use when calculating the extent tuple.
                If a list is given, it is converted into an array using
                :meth:`numpy.asarray`. If the array has the shape ``(m, 2)``, then there
                are ``m`` values in two dimensions (being the second dimension of
                ``data``). If the array has the shape ``(m,)`` or ``(m, 1)``, then
                ``dim`` must be set to the index of the dimension for which this data is
                associated.
            dim: Optional dimension index if a one-dimensional array is provided (i.e.,
                `X=0` and `Y=1`). If None, then ``data`` must have the shape ``(m, 2)``.
                Defaults to None.

        Raises:
            QiskitError: If the data is not two-dimensional and ``dim`` is not set.
            QiskitError: If the data does not contain one-dimensional values when
            ``dim`` is set.
            QiskitError: If ``dim`` is not an index for two-dimensions: i.e.,
            :math:`0\leq{}\text{dim}<2`.
        """
        data = np.asarray(data)
        if dim is None and (len(data.shape) != 2 or (len(data.shape) == 2 and data.shape[1] != 2)):
            raise QiskitError(
                "Data must contain two-dimensional values. Expected shape `(m,2)` but got "
                f"{data.shape} instead."
            )
        if dim is not None and (
            len(data.shape) > 2 or (len(data.shape) == 2 and data.shape[1] != 1)
        ):
            raise QiskitError(
                "Data must contain one-dimensional values if `dim` is set. Expected shape of `(m,1)` "
                f"or `(m,)` but got {data.shape} instead."
            )
        if dim is not None and dim >= 2:
            raise QiskitError(
                f"Dim must be a two-dimensional dimension index (0<=dim<2), got {dim} instead."
            )

        if dim is not None:
            _data_extent = [
                min(np.min(data), self._data_extent[dim, 0]),
                max(np.max(data), self._data_extent[dim, 1]),
            ]
            self._data_extent[dim] = _data_extent
        else:
            for i_dim in range(2):
                _data_extent = [
                    min(np.min(data[..., i_dim]), self._data_extent[i_dim, 0]),
                    max(np.max(data[..., i_dim]), self._data_extent[i_dim, 1]),
                ]
                self._data_extent[i_dim, :] = _data_extent

    @classmethod
    def _range(cls, extent: np.ndarray) -> np.ndarray:
        """Compute the range of the extent array.

        Args:
            extent: The extent array for the range.

        Returns:
            The array ``[x_range, y_range]`` where ``x_range`` and ``y_range`` are the
            ranges for their respective dimensions.
        """
        return np.diff(
            extent,
            axis=1,
        ).flatten()

    @classmethod
    def _midpoint(cls, extent: np.ndarray) -> np.ndarray:
        """Compute the midpoint of the extent array.

        Args:
            extent: The extent array for the midpoint.

        Returns:
            The array ``[x, y]`` where ``x`` and ``y`` are the midpoints for their
            respective dimensions.
        """
        return np.mean(
            extent,
            axis=1,
        )

    @classmethod
    def _extent_from_range_midpoint(
        cls, extent_range: np.ndarray, midpoint: np.ndarray
    ) -> np.ndarray:
        """Compute an extent array from range and midpoint arrays.

        Args:
            extent_range: The extent range to use.
            midpoint: The midpoint of the extent.

        Returns:
            An extent array with range and midpoint corresponding to ``extent_range``
            and ``midpoint``.
        """
        radii = extent_range.flatten() / 2
        new_extent = np.zeros((2, 2))
        new_extent[:, 0] = midpoint - radii
        new_extent[:, 1] = midpoint + radii
        return new_extent

    @classmethod
    def _extent_with_aspect_ratio(
        cls, extent: np.ndarray, target_aspect_ratio: Optional[float]
    ) -> np.ndarray:
        """Expand ``extent`` to have an aspect ratio defined by ``target_aspect_ratio``.

        This method extends ``extent`` along one of the two dimensions so that its
        aspect ratio is equal to ``target_aspect_ratio``. This is done by computing the
        ratio of the actual and target aspect ratios, computing the dimension along
        which to extend ``extent``, and recomputing the extent array based on scaled
        ranges to achieve the target aspect ratio. If the target ratio is ``None``,
        nothing is done. If extended, the region defined by the output array is always
        larger than the input array.

        Args:
            extent: The extent array to expand.
            target_aspect_ratio: Optional target aspect ratio, being
                :math:`\text{width}/\text{height}` for the extent array. If None,
                ``extent`` is not extended. Defaults to None.

        Returns:
            ``extent`` extended to have an aspect ratio defined by
            ``target_aspect_ratio``.
        """
        if target_aspect_ratio is None:
            return extent

        # Compute ratio coefficient
        _range = cls._range(extent)
        ratio = _range[0] / _range[1]
        ratio_coefficient = target_aspect_ratio / ratio

        _new_range = _range
        # Handle three cases of aspect ratios
        if ratio_coefficient == 1.0:
            return extent
        elif ratio_coefficient < 1:
            _new_range[1] = (1 / ratio_coefficient) * _range[1]
        else:
            _new_range[0] = ratio_coefficient * _range[0]

        return cls._extent_from_range_midpoint(_new_range, cls._midpoint(extent))

    @classmethod
    def _scaled_extent(cls, extent: np.ndarray, multiplier: float) -> np.ndarray:
        _range = cls._range(extent)
        midpoint = cls._midpoint(extent)
        new_range = multiplier * _range
        return cls._extent_from_range_midpoint(new_range, midpoint)

    def extent(self) -> ExtentTuple:
        """An extent array for the registered data, multiplier, and aspect ratio.

        Raises:
            QiskitError: If the resulting extent tuple is not finite. This can occur if
            no data was registered before calling :meth:`extent`.

        Returns:
            The extent tuple for the registered data, scaled by ``multiplier``, and then
            extended to achieve the set aspect ratio.
        """
        if not np.all(np.isfinite(self._data_extent)):
            is_infinite = np.argwhere(np.invert(np.isfinite(self._data_extent)))
            raise QiskitError(
                "Encountered non-finite extent. The following dimensions have non-finite "
                f"values: {is_infinite}."
            )

        return tuple(
            self._extent_with_aspect_ratio(
                self._scaled_extent(self._data_extent, self._multiplier), self._aspect_ratio
            ).flatten()
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Quantum Process Tomography experiment
"""

from typing import Union, Optional, List, Tuple, Sequence
import numpy as np
from qiskit.circuit import QuantumCircuit, Instruction, Clbit
from qiskit.providers.backend import Backend
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info import Choi, Operator, Statevector, DensityMatrix, partial_trace

from qiskit_experiments.exceptions import QiskitError
from .tomography_experiment import TomographyExperiment, TomographyAnalysis, BaseAnalysis
from .qpt_analysis import ProcessTomographyAnalysis
from . import basis


class ProcessTomography(TomographyExperiment):
    """An experiment to reconstruct the quantum channel from measurement data.

    # section: overview
        Quantum process tomography (QPT) is a method for experimentally
        reconstructing the quantum channel from measurement data.

        A QPT experiment prepares multiple input states, evolves them by the
        circuit, then performs multiple measurements in different measurement
        bases. The resulting measurement data is then post-processed by a
        tomography fitter to reconstruct the quantum channel.

    # section: note
        Performing full process tomography on an `N`-qubit circuit requires
        running :math:`4^N 3^N` measurement circuits when using the default
        preparation and measurement bases.

    # section: analysis_ref
        :class:`ProcessTomographyAnalysis`

    """

    def __init__(
        self,
        circuit: Union[QuantumCircuit, Instruction, BaseOperator],
        backend: Optional[Backend] = None,
        physical_qubits: Optional[Sequence[int]] = None,
        measurement_basis: basis.MeasurementBasis = basis.PauliMeasurementBasis(),
        measurement_indices: Optional[Sequence[int]] = None,
        preparation_basis: basis.PreparationBasis = basis.PauliPreparationBasis(),
        preparation_indices: Optional[Sequence[int]] = None,
        basis_indices: Optional[Sequence[Tuple[List[int], List[int]]]] = None,
        conditional_circuit_clbits: Union[bool, Sequence[int], Sequence[Clbit]] = False,
        analysis: Union[BaseAnalysis, None, str] = "default",
        target: Union[Statevector, DensityMatrix, None, str] = "default",
    ):
        """Initialize a quantum process tomography experiment.

        Args:
            circuit: the quantum process circuit. If not a quantum circuit
                it must be a class that can be appended to a quantum circuit.
            backend: The backend to run the experiment on.
            physical_qubits: Optional, the physical qubits for the initial state circuit.
                If None this will be qubits [0, N) for an N-qubit circuit.
            measurement_basis: Tomography basis for measurements. If not specified the
                default basis is the :class:`~basis.PauliMeasurementBasis`.
            measurement_indices: Optional, the `physical_qubits` indices to be measured.
                If None all circuit physical qubits will be measured.
            preparation_basis: Tomography basis for measurements. If not specified the
                default basis is the :class:`~basis.PauliPreparationBasis`.
            preparation_indices: Optional, the `physical_qubits` indices to be prepared.
                If None all circuit physical qubits will be prepared.
            basis_indices: Optional, a list of basis indices for generating partial
                tomography measurement data. Each item should be given as a pair of
                lists of preparation and measurement basis configurations
                ``([p[0], p[1], ...], [m[0], m[1], ...])``, where ``p[i]`` is the
                preparation basis index, and ``m[i]`` is the measurement basis index
                for qubit-i. If not specified full tomography for all indices of the
                preparation and measurement bases will be performed.
            conditional_circuit_clbits: Optional, the clbits in the source circuit to
                be conditioned on when reconstructing the channel. If True all circuit
                clbits will be conditioned on. Enabling this will return a list of
                reconstructed channel components conditional on the values of these clbit
                values.
            analysis: Optional, a custom analysis instance to use. If ``"default"``
                :class:`~.ProcessTomographyAnalysis` will be used. If None no analysis
                instance will be set.
            target: Optional, a custom quantum state target for computing the
                state fidelity of the fitted density matrix during analysis.
                If "default" the state will be inferred from the input circuit
                if it contains no classical instructions.
        """
        if analysis == "default":
            analysis = ProcessTomographyAnalysis()

        super().__init__(
            circuit,
            backend=backend,
            physical_qubits=physical_qubits,
            measurement_basis=measurement_basis,
            measurement_indices=measurement_indices,
            preparation_basis=preparation_basis,
            preparation_indices=preparation_indices,
            basis_indices=basis_indices,
            conditional_circuit_clbits=conditional_circuit_clbits,
            analysis=analysis,
        )

        # Set target quantum channel
        if isinstance(self.analysis, TomographyAnalysis):
            if target == "default":
                target = self._target_quantum_channel()
            self.analysis.set_options(target=target)

    def _target_quantum_channel(self) -> Union[Choi, Operator]:
        """Return the process tomography target"""
        # Check if circuit contains measure instructions
        # If so we cannot return target state
        circuit_ops = self._circuit.count_ops()
        if "measure" in circuit_ops:
            return None

        try:
            circuit = self._permute_circuit()
            if "reset" in circuit_ops or "kraus" in circuit_ops or "superop" in circuit_ops:
                channel = Choi(circuit)
            else:
                channel = Operator(circuit)
        except QiskitError:
            # Circuit couldn't be simulated
            return None

        total_qubits = self._circuit.num_qubits
        num_meas = total_qubits if not self._meas_indices else len(self._meas_indices)
        num_prep = total_qubits if not self._prep_indices else len(self._prep_indices)

        # If all qubits are prepared or measurement we are done
        if num_meas == total_qubits and num_prep == total_qubits:
            return channel

        # Convert channel to a state to project and trace out non-tomography
        # input and output qubits
        if isinstance(channel, Operator):
            chan_state = Statevector(np.ravel(channel, order="F"))
        else:
            chan_state = DensityMatrix(channel.data)

        # Get qargs for non measured and prepared subsystems
        non_meas_qargs = list(range(num_meas, total_qubits))
        non_prep_qargs = list(range(total_qubits + num_prep, 2 * total_qubits))

        # Project non-prepared subsystems on to the zero state
        if non_prep_qargs:
            proj0 = Operator([[1, 0], [0, 0]])
            for qarg in non_prep_qargs:
                chan_state = chan_state.evolve(proj0, [qarg])

        # Trace out indices to remove
        tr_qargs = non_meas_qargs + non_prep_qargs
        chan_state = partial_trace(chan_state, tr_qargs)
        channel = Choi(chan_state.data, input_dims=[2] * num_prep, output_dims=[2] * num_meas)
        return channel



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Cross resonance Hamiltonian tomography experiment analysis."""

from typing import List, Dict
import numpy as np

from qiskit.utils.deprecation import deprecate_func

import qiskit_experiments.curve_analysis as curve
from qiskit_experiments.framework import AnalysisResultData
from qiskit_experiments.visualization import PlotStyle


class CrossResonanceHamiltonianAnalysis(curve.CompositeCurveAnalysis):
    r"""A class to analyze cross resonance Hamiltonian tomography experiment.

    # section: fit_model

        This analysis performs :class:`.BlochTrajectoryAnalysis` on the target qubit
        with the control qubit states in :math:`\in \{ |0\rangle, |1\rangle \}`.

        Based on the fit result, cross resonance Hamiltonian coefficients can be determined by

        .. math::

            ZX &= \frac{p_{x, |0\rangle} - p_{x, |1\rangle}}{2}, \\
            ZY &= \frac{p_{y, |0\rangle} - p_{y, |1\rangle}}{2}, \\
            ZZ &= \frac{p_{z, |0\rangle} - p_{z, |1\rangle}}{2}, \\
            IX &= \frac{p_{x, |0\rangle} + p_{x, |1\rangle}}{2}, \\
            IY &= \frac{p_{y, |0\rangle} + p_{y, |1\rangle}}{2}, \\
            IZ &= \frac{p_{z, |0\rangle} + p_{z, |1\rangle}}{2},

        where :math:`p_{\beta, |j\rangle}` is a fit parameter of :class:`.BlochTrajectoryAnalysis`
        for the projection axis :math:`\beta` with the control qubit state :math:`|j\rangle`.

    """

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments and related classses "
            "involving pulse gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(self):
        analyses = []
        for control_state in (0, 1):
            analysis = curve.BlochTrajectoryAnalysis(name=f"ctrl{control_state}")
            analysis.set_options(filter_data={"control_state": control_state})
            analyses.append(analysis)

        super().__init__(analyses=analyses)

    @classmethod
    def _default_options(cls):
        """Return the default analysis options."""
        default_options = super()._default_options()
        default_options.plotter.set_options(
            subplots=(3, 1),
            style=PlotStyle(
                {
                    "figsize": (8, 10),
                    "legend_loc": "lower right",
                    "textbox_rel_pos": (0.28, -0.10),
                }
            ),
        )
        default_options.plotter.set_figure_options(
            xlabel="Flat top width",
            ylabel=[
                r"$\langle$X(t)$\rangle$",
                r"$\langle$Y(t)$\rangle$",
                r"$\langle$Z(t)$\rangle$",
            ],
            xval_unit="s",
            ylim=(-1, 1),
            series_params={
                "x_ctrl0": {
                    "canvas": 0,
                    "color": "blue",
                    "label": "X (ctrl0)",
                    "symbol": "o",
                },
                "y_ctrl0": {
                    "canvas": 1,
                    "color": "blue",
                    "label": "Y (ctrl0)",
                    "symbol": "o",
                },
                "z_ctrl0": {
                    "canvas": 2,
                    "color": "blue",
                    "label": "Z (ctrl0)",
                    "symbol": "o",
                },
                "x_ctrl1": {
                    "canvas": 0,
                    "color": "red",
                    "label": "X (ctrl1)",
                    "symbol": "^",
                },
                "y_ctrl1": {
                    "canvas": 1,
                    "color": "red",
                    "label": "Y (ctrl1)",
                    "symbol": "^",
                },
                "z_ctrl1": {
                    "canvas": 2,
                    "color": "red",
                    "label": "Z (ctrl1)",
                    "symbol": "^",
                },
            },
        )

        return default_options

    def _create_analysis_results(
        self,
        fit_data: Dict[str, curve.CurveFitResult],
        quality: str,
        **metadata,
    ) -> List[AnalysisResultData]:
        outcomes = []

        for control in ("z", "i"):
            for target in ("x", "y", "z"):
                p0_val = fit_data["ctrl0"].ufloat_params[f"p{target}"]
                p1_val = fit_data["ctrl1"].ufloat_params[f"p{target}"]

                if control == "z":
                    coef_val = 0.5 * (p0_val - p1_val) / (2 * np.pi)
                else:
                    coef_val = 0.5 * (p0_val + p1_val) / (2 * np.pi)

                outcomes.append(
                    AnalysisResultData(
                        name=f"omega_{control}{target}",
                        value=coef_val,
                        quality=quality,
                        extra={
                            "unit": "Hz",
                            **metadata,
                        },
                    )
                )

        return outcomes



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Quantum Tomography experiment
"""

from typing import Union, Optional, Iterable, List, Tuple, Sequence
from itertools import product
from qiskit.circuit import QuantumCircuit, Instruction, ClassicalRegister, Clbit
from qiskit.circuit.library import Permutation
from qiskit.providers.backend import Backend
from qiskit.quantum_info.operators.base_operator import BaseOperator

from qiskit_experiments.exceptions import QiskitError
from qiskit_experiments.framework import BaseExperiment, BaseAnalysis, Options
from .basis import PreparationBasis, MeasurementBasis
from .tomography_analysis import TomographyAnalysis


class TomographyExperiment(BaseExperiment):
    """Base experiment for quantum state and process tomography.

    # section: analysis_ref
        :class:`TomographyAnalysis`
    """

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            basis_indices (Iterable[Tuple[List[int], List[int]]]): The basis elements to be measured.
                If None All basis elements will be measured.

        """
        options = super()._default_experiment_options()
        options.basis_indices = None
        return options

    def __init__(
        self,
        circuit: Union[QuantumCircuit, Instruction, BaseOperator],
        backend: Optional[Backend] = None,
        physical_qubits: Optional[Sequence[int]] = None,
        measurement_basis: Optional[MeasurementBasis] = None,
        measurement_indices: Optional[Sequence[int]] = None,
        preparation_basis: Optional[PreparationBasis] = None,
        preparation_indices: Optional[Sequence[int]] = None,
        conditional_circuit_clbits: Union[bool, Sequence[int], Sequence[Clbit]] = False,
        basis_indices: Optional[Iterable[Tuple[List[int], List[int]]]] = None,
        analysis: Union[BaseAnalysis, None, str] = "default",
    ):
        """Initialize a tomography experiment.

        Args:
            circuit: the quantum process circuit. If not a quantum circuit
                it must be a class that can be appended to a quantum circuit.
            backend: The backend to run the experiment on.
            physical_qubits: Optional, the physical qubits for the initial state circuit.
                If None this will be qubits [0, N) for an N-qubit circuit.
            measurement_basis: Tomography basis for measurements. If set to None
                no tomography measurements will be performed.
            measurement_indices: Optional, the `physical_qubits` indices to be
                measured as specified by the `measurement_basis`. If None all
                circuit physical qubits will be measured.
            preparation_basis: Tomography basis for measurements. If set to None
                no tomography preparations will be performed.
            preparation_indices: Optional, the `physical_qubits` indices to be
                prepared as specified by the `preparation_basis`. If None all
                circuit physical qubits will be prepared.
            basis_indices: Optional, the basis elements to be measured. If None
                All basis elements will be measured.
            conditional_circuit_clbits: Specify any clbits in the input
                circuit to treat as conditioning bits for conditional tomography.
                If set to True all circuit clbits will be treated as conditional.
                If False all circuit clbits will be marginalized over (Default: False).
            analysis: Optional, a custom analysis instance to use. If ``"default"``
                :class:`~.TomographyAnalysis` will be used. If None no analysis
                instance will be set.

        Raises:
            QiskitError: If input params are invalid.
        """
        # Initialize BaseExperiment
        if physical_qubits is None:
            physical_qubits = tuple(range(circuit.num_qubits))
        if analysis == "default":
            analysis = TomographyAnalysis()
        super().__init__(physical_qubits, analysis=analysis, backend=backend)

        # Get the target tomography circuit
        if isinstance(circuit, QuantumCircuit):
            target_circuit = circuit
        else:
            # Convert input to a circuit
            num_qubits = circuit.num_qubits
            target_circuit = QuantumCircuit(num_qubits)
            target_circuit.append(circuit, range(num_qubits))
        self._circuit = target_circuit

        self._cond_clbits = None
        if conditional_circuit_clbits is True:
            conditional_circuit_clbits = self._circuit.clbits
        if conditional_circuit_clbits:
            cond_clbits = []
            for i in conditional_circuit_clbits:
                if isinstance(i, Clbit):
                    cond_clbits.append(self._circuit.find_bit(i).index)
                elif i < self._circuit.num_clbits:
                    cond_clbits.append(i)
                else:
                    raise QiskitError(
                        f"Circuit {self._circuit.name} does not contain conditional clbit {i}"
                    )
            self._cond_clbits = cond_clbits

        # Measurement basis and qubits
        self._meas_circ_basis = measurement_basis
        if measurement_indices:
            # Convert logical qubits to physical qubits
            self._meas_indices = tuple(measurement_indices)
            self._meas_physical_qubits = tuple(self.physical_qubits[i] for i in self._meas_indices)
            for qubit in self._meas_indices:
                if qubit not in range(self.num_qubits):
                    raise QiskitError(
                        f"measurement qubit ({qubit}) is outside the range"
                        f" of circuit qubits [0, {self.num_qubits})."
                    )
        elif measurement_basis:
            self._meas_indices = tuple(range(self.num_qubits))
            self._meas_physical_qubits = self.physical_qubits
        else:
            self._meas_indices = tuple()
            self._meas_physical_qubits = tuple()

        # Preparation basis and qubits
        self._prep_circ_basis = preparation_basis
        if preparation_indices:
            self._prep_indices = tuple(preparation_indices)
            self._prep_physical_qubits = tuple(self.physical_qubits[i] for i in self._prep_indices)
            for qubit in self._prep_indices:
                if qubit not in range(self.num_qubits):
                    raise QiskitError(
                        f"preparation qubit ({qubit}) is outside the range"
                        f" of circuit qubits [0, {self.num_qubits})."
                    )
        elif preparation_basis:
            self._prep_indices = tuple(range(self.num_qubits))
            self._prep_physical_qubits = self.physical_qubits
        else:
            self._prep_indices = tuple()
            self._prep_physical_qubits = tuple()

        # Configure experiment options
        if basis_indices:
            self.set_experiment_options(basis_indices=basis_indices)

        # Configure analysis basis options
        if isinstance(self.analysis, TomographyAnalysis):
            analysis_options = {}
            if measurement_basis:
                analysis_options["measurement_basis"] = measurement_basis
                analysis_options["measurement_qubits"] = self._meas_physical_qubits
            if preparation_basis:
                analysis_options["preparation_basis"] = preparation_basis
                analysis_options["preparation_qubits"] = self._prep_physical_qubits
            if conditional_circuit_clbits:
                analysis_options["conditional_circuit_clbits"] = self._cond_clbits
            self.analysis.set_options(**analysis_options)

    def circuits(self):
        circ_qubits = self._circuit.qubits
        circ_clbits = self._circuit.clbits
        meas_creg = ClassicalRegister((len(self._meas_indices)), name="c_tomo")
        template = QuantumCircuit(
            *self._circuit.qregs, *self._circuit.cregs, meas_creg, name=f"{self._type}"
        )
        if self._circuit.metadata:
            template.metadata = self._circuit.metadata.copy()
        else:
            template.metadata = {}
        meas_clbits = [template.find_bit(i).index for i in meas_creg]

        # Build circuits
        circuits = []
        for prep_element, meas_element in self._basis_indices():
            name = template.name
            metadata = {"clbits": meas_clbits, "cond_clbits": self._cond_clbits}
            if meas_element:
                name += f"_{meas_element}"
                metadata["m_idx"] = list(meas_element)
            if prep_element:
                name += f"_{prep_element}"
                metadata["p_idx"] = list(prep_element)

            circ = template.copy(name=name)

            if prep_element:
                # Add tomography preparation
                prep_circ = self._prep_circ_basis.circuit(prep_element, self._prep_physical_qubits)
                circ.compose(prep_circ, self._prep_indices, inplace=True)
                circ.barrier(*self._prep_indices)

            # Add target circuit
            # Have to use compose since circuit.to_instruction has a bug
            # when circuit contains classical registers and conditionals
            circ.compose(self._circuit, circ_qubits, circ_clbits, inplace=True)

            # Add tomography measurement
            if meas_element:
                meas_circ = self._meas_circ_basis.circuit(meas_element, self._meas_physical_qubits)
                circ.barrier(*self._meas_indices)
                circ.compose(meas_circ, self._meas_indices, meas_clbits, inplace=True)

            # Add metadata
            circ.metadata.update(**metadata)
            circuits.append(circ)
        return circuits

    def _metadata(self):
        metadata = super()._metadata()
        if self._meas_physical_qubits:
            metadata["m_qubits"] = list(self._meas_physical_qubits)
        if self._prep_physical_qubits:
            metadata["p_qubits"] = list(self._prep_physical_qubits)
        return metadata

    def _basis_indices(self):
        """Return list of basis element indices"""
        basis_indices = self.experiment_options.basis_indices
        if basis_indices is not None:
            return basis_indices

        if self._meas_circ_basis:
            meas_shape = self._meas_circ_basis.index_shape(self._meas_physical_qubits)
            ranges = [range(i) for i in meas_shape]
            meas_elements = product(*ranges)
        else:
            meas_elements = [None]
        if self._prep_circ_basis:
            prep_shape = self._prep_circ_basis.index_shape(self._prep_physical_qubits)
            prep_elements = product(*[range(i) for i in prep_shape])
        else:
            prep_elements = [None]
        return product(prep_elements, meas_elements)

    def _permute_circuit(self) -> QuantumCircuit:
        """Permute circuit qubits.

        This permutes the circuit so that the specified preparation and measurement
        qubits correspond to input and output qubits [0, ..., N-1] and [0, ..., M-1]
        respectively for the returned circuit.
        """
        default_range = tuple(range(self.num_qubits))
        permute_meas = self._meas_indices and self._meas_indices != default_range
        permute_prep = self._prep_indices and self._prep_indices != default_range
        if not permute_meas and not permute_prep:
            return self._circuit

        total_qubits = self._circuit.num_qubits
        total_clbits = self._circuit.num_clbits
        if total_clbits:
            perm_circ = QuantumCircuit(total_qubits, total_clbits)
        else:
            perm_circ = QuantumCircuit(total_qubits)

        # Apply permutation to put prep qubits as [0, ..., M-1]
        if self._prep_indices:
            prep_qargs = list(self._prep_indices)
            if len(self._prep_indices) != total_qubits:
                prep_qargs += [i for i in range(total_qubits) if i not in self._prep_indices]
            perm_circ.append(Permutation(total_qubits, prep_qargs).inverse(), range(total_qubits))

        # Apply original circuit
        if total_clbits:
            perm_circ = perm_circ.compose(self._circuit, range(total_qubits), range(total_clbits))
        else:
            perm_circ = perm_circ.compose(self._circuit, range(total_qubits))

        # Apply permutation to put meas qubits as [0, ..., M-1]
        if self._meas_indices:
            meas_qargs = list(self._meas_indices)
            if len(self._meas_indices) != total_qubits:
                meas_qargs += [i for i in range(total_qubits) if i not in self._meas_indices]
            perm_circ.append(Permutation(total_qubits, meas_qargs), range(total_qubits))

        return perm_circ



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Base analysis class.
"""
from abc import ABC, abstractmethod
import copy
from collections import OrderedDict
from datetime import datetime
from typing import List, Tuple, Union, Dict
import warnings

from dateutil import tz

from qiskit_experiments.database_service.device_component import Qubit
from qiskit_experiments.framework import Options
from qiskit_experiments.framework.containers.artifact_data import ArtifactData
from qiskit_experiments.framework.store_init_args import StoreInitArgs
from qiskit_experiments.framework.experiment_data import ExperimentData
from qiskit_experiments.framework.containers import FigureData, FigureType
from qiskit_experiments.framework.configs import AnalysisConfig
from qiskit_experiments.framework.analysis_result_data import AnalysisResultData, as_table_element


class BaseAnalysis(ABC, StoreInitArgs):
    """Abstract base class for analyzing Experiment data.

    The data produced by experiments (i.e. subclasses of BaseExperiment)
    are analyzed with subclasses of BaseAnalysis. The analysis is
    typically run after the data has been gathered by the experiment.
    For example, an analysis may perform some data processing of the
    measured data and a fit to a function to extract a parameter.

    Analysis subclasses must implement the abstract method `_run_analysis`.
    This method should not have side-effects on the analysis class itself
    since it could potentially be called asynchronously in multiple threads.
    Any configurable option values should be specified in the `_default_options`
    class method. These values can be overriden by a user by calling the
    `set_options` method or for a single-run can be specified by passing kwarg
    options to the :meth:`run` method.
    """

    def __init__(self):
        """Initialize the analysis object."""
        # Analysis options
        self._options = self._default_options()

        # Store keys of non-default options
        self._set_options = set()

    def config(self) -> AnalysisConfig:
        """Return the config dataclass for this analysis"""
        args = tuple(getattr(self, "__init_args__", OrderedDict()).values())
        kwargs = dict(getattr(self, "__init_kwargs__", OrderedDict()))
        # Only store non-default valued options
        options = dict((key, getattr(self._options, key)) for key in self._set_options)
        return AnalysisConfig(
            cls=type(self),
            args=args,
            kwargs=kwargs,
            options=options,
        )

    @classmethod
    def from_config(cls, config: Union[AnalysisConfig, Dict]) -> "BaseAnalysis":
        """Initialize an analysis class from analysis config"""
        if isinstance(config, dict):
            config = AnalysisConfig(**config)
        ret = cls(*config.args, **config.kwargs)
        if config.options:
            ret.set_options(**config.options)
        return ret

    def copy(self) -> "BaseAnalysis":
        """Return a copy of the analysis"""
        # We want to avoid a deep copy be default for performance so we
        # need to also copy the Options structures so that if they are
        # updated on the copy they don't effect the original.
        ret = copy.copy(self)
        ret._options = copy.copy(self._options)
        ret._set_options = copy.copy(self._set_options)
        return ret

    @classmethod
    def _default_options(cls) -> Options:
        """Default analysis options common to all analyses.

        Analysis Options:
            figure_names (str or List[str]): Identifier of figures that appear in the
                experiment data to sort figures by name.
        """
        options = Options()
        # figure names can be set for each analysis by calling
        # experiment_obj.analysis.set_options(figure_names=FIGURE_NAMES)
        options.figure_names = None
        return options

    @property
    def options(self) -> Options:
        """Return the analysis options for :meth:`run` method."""
        return self._options

    def set_options(self, **fields):
        """Set the analysis options for :meth:`run` method.

        Args:
            fields: The fields to update the options
        """
        self._options.update_options(**fields)
        self._set_options = self._set_options.union(fields)

    def run(
        self,
        experiment_data: ExperimentData,
        replace_results: bool = False,
        **options,
    ) -> ExperimentData:
        """Run analysis and update ExperimentData with analysis result.

        Args:
            experiment_data: the experiment data to analyze.
            replace_results: If True clear any existing analysis results, figures,
                             and artifacts in the experiment data and replace with
                             new results. See note for additional information.
            options: additional analysis options. See class documentation for
                     supported options.

        Returns:
            An experiment data object containing analysis results, figures, and artifacts.

        Raises:
            QiskitError: If experiment_data container is not valid for analysis.

        .. note::
            **Updating Results**

            If analysis is run with ``replace_results=True`` then any analysis results,
            figures, and artifacts in the experiment data will be cleared and replaced with the
            new analysis results. Saving this experiment data will replace any
            previously saved data in a database service using the same experiment ID.

            If analysis is run with ``replace_results=False`` and the experiment data
            being analyzed has already been saved to a database service, or already
            contains analysis results or figures, a copy with a unique experiment ID
            will be returned containing only the new analysis results and figures.
            This data can then be saved as its own experiment to a database service.
        """
        # Make a new copy of experiment data if not updating results
        if not replace_results and _requires_copy(experiment_data):
            experiment_data = experiment_data.copy()

        # Set Analysis options
        if not options:
            analysis = self
        else:
            analysis = self.copy()
            analysis.set_options(**options)

        def run_analysis(expdata: ExperimentData):
            # Clearing previous analysis data
            experiment_data._clear_results()

            if not expdata.data():
                warnings.warn("ExperimentData object data is empty.\n")

            # Making new analysis
            results, figures = analysis._run_analysis(expdata)

            if results:
                for result in results:
                    if isinstance(result, AnalysisResultData):
                        # Populate missing data fields
                        if not result.experiment_id:
                            result.experiment_id = expdata.experiment_id
                        if not result.experiment:
                            result.experiment = expdata.experiment_type
                        if not result.device_components:
                            result.device_components = analysis._get_experiment_components(expdata)
                        if not result.backend:
                            result.backend = expdata.backend_name
                        if not result.created_time:
                            result.created_time = datetime.now(tz.tzlocal())
                        if not result.run_time:
                            result.run_time = expdata.running_time

                        # To canonical kwargs to add to the analysis table.
                        table_format = as_table_element(result)

                        # Remove result_id to make sure the id is unique in the scope of the container.
                        # This will let the container generate a unique id.
                        del table_format["result_id"]

                        expdata.add_analysis_results(**table_format)
                    elif isinstance(result, ArtifactData):
                        if not result.experiment_id:
                            result.experiment_id = expdata.experiment_id
                        if not result.device_components:
                            result.device_components = analysis._get_experiment_components(expdata)
                        if not result.experiment:
                            result.experiment = expdata.experiment_type
                        expdata.add_artifacts(result)
                    else:
                        raise TypeError(
                            f"Invalid object type {result.__class__.__name__} for analysis results. "
                            "This data cannot be stored in the experiment data."
                        )

            if figures:
                figure_to_add = []
                for figure in figures:
                    if not isinstance(figure, FigureData):
                        qubits_repr = "_".join(
                            map(str, expdata.metadata.get("device_components", [])[:5])
                        )
                        short_id = expdata.experiment_id[:8]
                        figure = FigureData(
                            figure=figure,
                            name=f"{expdata.experiment_type}_{qubits_repr}_{short_id}.svg",
                        )
                    figure_to_add.append(figure)
                expdata.add_figures(figure_to_add, figure_names=analysis.options.figure_names)

        experiment_data.add_analysis_callback(run_analysis)

        return experiment_data

    def _get_experiment_components(self, experiment_data: ExperimentData):
        """Subclasses may override this method to specify the experiment components."""
        if "physical_qubits" in experiment_data.metadata:
            experiment_components = [
                Qubit(qubit) for qubit in experiment_data.metadata["physical_qubits"]
            ]
        else:
            experiment_components = []

        return experiment_components

    @abstractmethod
    def _run_analysis(
        self,
        experiment_data: ExperimentData,
    ) -> Tuple[List[Union[AnalysisResultData, ArtifactData]], List[FigureType]]:
        """Run analysis on circuit data.

        Args:
            experiment_data: the experiment data to analyze.

        Returns:
            A pair ``(analysis_results, figures)`` where ``analysis_results``
            is a list of :class:`.AnalysisResultData` objects, and ``figures``
            is a list of any figures for the experiment.

        Raises:
            AnalysisError: If the analysis fails.
        """
        # NOTE: passing kwarg options to _run_analysis should be removed once
        pass

    def __json_encode__(self):
        return self.config()

    @classmethod
    def __json_decode__(cls, value):
        return cls.from_config(value)


def _requires_copy(experiment_data) -> bool:
    """Return True if a copy of the experiment data should be made."""
    # If data is from DB or contains analysis results it should be copied
    if (
        experiment_data._created_in_db
        or experiment_data._analysis_results
        or experiment_data._figures
    ):
        return True

    # Check child data:
    if hasattr(experiment_data, "_child_data"):
        for subdata in experiment_data._child_data.values():
            if _requires_copy(subdata):
                return True

    # No Copy required
    return False



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Tphi Analysis class.
"""

from typing import List, Tuple

from qiskit_experiments.framework import ExperimentData, AnalysisResultData
from qiskit_experiments.framework.composite.composite_analysis import CompositeAnalysis
from qiskit_experiments.library.characterization.analysis import (
    T1Analysis,
    T2HahnAnalysis,
    T2RamseyAnalysis,
)
from qiskit_experiments.exceptions import QiskitError


class TphiAnalysis(CompositeAnalysis):
    r"""A class to analyze :math:`T_\phi` experiments.

    # section: see_also
        * :py:class:`qiskit_experiments.library.characterization.analysis.T1Analysis`
        * :py:class:`qiskit_experiments.library.characterization.analysis.T2HahnAnalysis`
        * :py:class:`qiskit_experiments.library.characterization.analysis.T2RamseyAnalysis`

    """

    def __init__(self, analyses=None):
        if analyses is None:
            analyses = [T1Analysis(), T2HahnAnalysis()]

        # Validate analyses kwarg
        if (
            len(analyses) != 2
            or not isinstance(analyses[0], T1Analysis)
            or not isinstance(analyses[1], (T2RamseyAnalysis, T2HahnAnalysis))
        ):
            raise QiskitError(
                "Invalid component analyses for Tphi, analyses must be a pair of "
                "T1Analysis and T2HahnAnalysis or T2RamseyAnalysis instances."
            )
        super().__init__(analyses, flatten_results=True)

    def _run_analysis(
        self, experiment_data: ExperimentData
    ) -> Tuple[List[AnalysisResultData], List["matplotlib.figure.Figure"]]:
        r"""Run analysis for :math:`T_\phi` experiment.
        It invokes CompositeAnalysis._run_analysis that will invoke
        _run_analysis for the two sub-experiments.
        Based on the results, it computes the result for :math:`T_phi`.
        """

        # Run composite analysis and extract T1 and T2 results
        analysis_results, figures = super()._run_analysis(experiment_data)
        t1_result = next(filter(lambda res: res.name == "T1", analysis_results))
        t2_result = next(filter(lambda res: res.name in {"T2star", "T2"}, analysis_results))

        # Calculate Tphi from T1 and T2
        tphi = 1 / (1 / t2_result.value - 1 / (2 * t1_result.value))
        quality_tphi = (
            "good" if (t1_result.quality == "good" and t2_result.quality == "good") else "bad"
        )
        tphi_result = AnalysisResultData(
            name="T_phi",
            value=tphi,
            chisq=None,
            quality=quality_tphi,
            extra={"unit": "s"},
        )

        # Return combined results
        analysis_results = [tphi_result] + analysis_results
        return analysis_results, figures



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Analyze oscillating data such as a Rabi amplitude experiment."""

from typing import List, Union, Optional

import lmfit
import numpy as np

import qiskit_experiments.curve_analysis as curve


class OscillationAnalysis(curve.CurveAnalysis):
    r"""Oscillation analysis class based on a fit of the data to a cosine function.

    # section: fit_model

        Analyse oscillating data by fitting it to a cosine function

        .. math::

            y = {\rm amp} \cos\left(2 \pi\cdot {\rm freq}\cdot x + {\rm phase}\right) + {\rm base}

    # section: fit_parameters
        defpar \rm amp:
            desc: Amplitude of the oscillation.
            init_guess: Calculated by :func:`~qiskit_experiments.curve_analysis.guess.max_height`.
            bounds: [-2, 2] scaled to the maximum signal value.

        defpar \rm base:
            desc: Base line.
            init_guess: Calculated by :func:`~qiskit_experiments.curve_analysis.\
            guess.constant_sinusoidal_offset`.
            bounds: [-1, 1] scaled to the maximum signal value.

        defpar \rm freq:
            desc: Frequency of the oscillation. This is the fit parameter of interest.
            init_guess: Calculated by :func:`~qiskit_experiments.curve_analysis.\
            guess.frequency`.
            bounds: [0, inf].

        defpar \rm phase:
            desc: Phase of the oscillation.
            init_guess: Zero.
            bounds: [-pi, pi].
    """

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        super().__init__(
            models=[
                lmfit.models.ExpressionModel(
                    expr="amp * cos(2 * pi * freq * x + phase) + base",
                    name="cos",
                )
            ],
            name=name,
        )

    def _generate_fit_guesses(
        self,
        user_opt: curve.FitOptions,
        curve_data: curve.ScatterTable,
    ) -> Union[curve.FitOptions, List[curve.FitOptions]]:
        """Create algorithmic initial fit guess from analysis options and curve data.

        Args:
            user_opt: Fit options filled with user provided guess and bounds.
            curve_data: Formatted data collection to fit.

        Returns:
            List of fit options that are passed to the fitter function.
        """
        max_abs_y, _ = curve.guess.max_height(curve_data.y, absolute=True)

        user_opt.bounds.set_if_empty(
            amp=(-2 * max_abs_y, 2 * max_abs_y),
            freq=(0, np.inf),
            phase=(-np.pi, np.pi),
            base=(-max_abs_y, max_abs_y),
        )
        user_opt.p0.set_if_empty(
            freq=curve.guess.frequency(curve_data.x, curve_data.y),
            base=curve.guess.constant_sinusoidal_offset(curve_data.y),
        )
        user_opt.p0.set_if_empty(
            amp=curve.guess.max_height(curve_data.y - user_opt.p0["base"], absolute=True)[0],
        )

        options = []
        for phase_guess in np.linspace(0, np.pi, 5):
            new_opt = user_opt.copy()
            new_opt.p0.set_if_empty(phase=phase_guess)
            options.append(new_opt)

        return options

    def _evaluate_quality(self, fit_data: curve.CurveFitResult) -> Union[str, None]:
        """Algorithmic criteria for whether the fit is good or bad.

        A good fit has:
            - a reduced chi-squared lower than three and greater than zero,
            - more than a quarter of a full period,
            - less than 10 full periods, and
            - an error on the fit frequency lower than the fit frequency.
        """
        fit_freq = fit_data.ufloat_params["freq"]

        criteria = [
            0 < fit_data.reduced_chisq < 3,
            1.0 / 4.0 < fit_freq.nominal_value < 10.0,
            curve.utils.is_error_not_significant(fit_freq),
        ]

        if all(criteria):
            return "good"

        return "bad"


class DampedOscillationAnalysis(curve.CurveAnalysis):
    r"""A class to analyze general exponential decay curve with sinusoidal oscillation.

    # section: fit_model
        This class is based on the fit model of sinusoidal signal with exponential decay.
        This model is often used for the oscillation signal in the dissipative system.

        .. math::

            F(x) = {\rm amp} \cdot e^{-x/\tau}
                \cos(2\pi \cdot {\rm freq} \cdot t + \phi) + {\rm base}

    # section: fit_parameters

        defpar \rm amp:
            desc: Amplitude. Height of the decay curve.
            init_guess: 0.5
            bounds: [0, 1.5],

        defpar \rm base:
            desc: Offset. Base line of the decay curve.
            init_guess: Determined by :func:`~qiskit_experiments.curve_analysis.\
                guess.constant_sinusoidal_offset`
            bounds: [0, 1.5]

        defpar \tau:
            desc: Represents the rate of decay.
            init_guess: Determined by :func:`~qiskit_experiments.curve_analysis.\
                guess.oscillation_exp_decay`
            bounds: [0, None]

        defpar \rm freq:
            desc: Oscillation frequency.
            init_guess: Determined by :func:`~qiskit_experiments.curve_analysis.guess.frequency`
            bounds: [0, 10 freq]

        defpar \phi:
            desc: Phase. Relative shift of the sinusoidal function from the origin.
            init_guess: Set multiple guesses within [-pi, pi]
            bounds: [-pi, pi]
    """

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        super().__init__(
            models=[
                lmfit.models.ExpressionModel(
                    expr="amp * exp(-x / tau) * cos(2 * pi * freq * x + phi) + base",
                    name="cos_decay",
                )
            ],
            name=name,
        )

    def _generate_fit_guesses(
        self,
        user_opt: curve.FitOptions,
        curve_data: curve.ScatterTable,
    ) -> Union[curve.FitOptions, List[curve.FitOptions]]:
        """Create algorithmic initial fit guess from analysis options and curve data.

        Args:
            user_opt: Fit options filled with user provided guess and bounds.
            curve_data: Formatted data collection to fit.

        Returns:
            List of fit options that are passed to the fitter function.
        """
        user_opt.p0.set_if_empty(
            amp=0.5,
            base=curve.guess.constant_sinusoidal_offset(curve_data.y),
        )

        # frequency resolution of this scan
        df = 1 / ((curve_data.x[1] - curve_data.x[0]) * len(curve_data.x))

        if user_opt.p0["freq"] is not None:
            # If freq guess is provided
            freq_guess = user_opt.p0["freq"]

            freqs = [freq_guess]
        else:
            freq_guess = curve.guess.frequency(curve_data.x, curve_data.y - user_opt.p0["base"])

            # The FFT might be up to 1/2 bin off
            if freq_guess > df:
                freqs = [freq_guess - df, freq_guess, freq_guess + df]
            else:
                freqs = [0.0, freq_guess]

        # Set guess for decay parameter based on estimated frequency
        if freq_guess > df:
            alpha = curve.guess.oscillation_exp_decay(
                curve_data.x, curve_data.y - user_opt.p0["base"], freq_guess=freq_guess
            )
        else:
            # Very low frequency. Assume standard exponential decay
            alpha = curve.guess.exp_decay(curve_data.x, curve_data.y)

        if alpha != 0.0:
            user_opt.p0.set_if_empty(tau=-1 / alpha)
        else:
            # Likely there is no slope. Cannot fit constant line with this model.
            # Set some large enough number against to the scan range.
            user_opt.p0.set_if_empty(tau=100 * np.max(curve_data.x))

        user_opt.bounds.set_if_empty(
            amp=[0, 1.5],
            base=[0, 1.5],
            tau=(0, np.inf),
            freq=(0, 10 * freq_guess),
            phi=(-np.pi, np.pi),
        )

        # more robust estimation
        options = []
        for freq in freqs:
            for phi in np.linspace(-np.pi, np.pi, 5)[:-1]:
                new_opt = user_opt.copy()
                new_opt.p0.set_if_empty(freq=freq, phi=phi)
                options.append(new_opt)

        return options

    def _evaluate_quality(self, fit_data: curve.CurveFitResult) -> Union[str, None]:
        """Algorithmic criteria for whether the fit is good or bad.

        A good fit has:
            - a reduced chi-squared lower than three and greater than zero
            - relative error of tau is less than its value
            - relative error of freq is less than its value
        """
        tau = fit_data.ufloat_params["tau"]
        freq = fit_data.ufloat_params["freq"]

        criteria = [
            0 < fit_data.reduced_chisq < 3,
            curve.utils.is_error_not_significant(tau),
            curve.utils.is_error_not_significant(freq),
        ]

        if all(criteria):
            return "good"

        return "bad"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Rough drag experiment."""

import warnings
from typing import Iterable, List, Optional, Sequence
import numpy as np

from qiskit import QuantumCircuit
from qiskit.circuit import Gate
from qiskit.exceptions import QiskitError
from qiskit.providers.backend import Backend
from qiskit.pulse import ScheduleBlock
from qiskit.utils.deprecation import deprecate_func

from qiskit_experiments.framework import BaseExperiment, Options
from qiskit_experiments.framework.restless_mixin import RestlessMixin
from qiskit_experiments.library.characterization.analysis import DragCalAnalysis


class RoughDrag(BaseExperiment, RestlessMixin):
    r"""An experiment that scans the DRAG parameter to find the optimal value.

    # section: overview

        A Derivative Removal by Adiabatic Gate (DRAG) pulse is designed to minimize phase
        errors and leakage resulting from the presence of a neighbouring transition. DRAG
        is a standard pulse with an additional derivative component. The optimal value of
        the DRAG parameter, :math:`\beta`, is chosen to primarily minimize phase errors
        resulting from the AC Stark shift and potentially leakage errors. The DRAG pulse is

        .. math::

            f(t) = \Omega(t) + 1j \beta d/dt \Omega(t)

        Here, :math:`\Omega` is the envelop of the in-phase component of the pulse and
        :math:`\beta` is the strength of the quadrature which we refer to as the DRAG
        parameter and seek to calibrate in this experiment. The DRAG calibration will run
        several series of circuits. In a given circuit a Rp(β) - Rm(β) block is repeated
        :math:`N` times. Here, Rp is a rotation with a positive angle and Rm is the same rotation
        with a native angle and is implemented by the gate sequence Rz(π) - Rp(β) - Rz(π) where
        the Z rotations are virtual. As example the circuit of a single repetition, i.e.
        :math:`N=1`, is shown below.

        .. parsed-literal::

                       ┌───────┐┌───────┐┌───────┐┌───────┐ ░ ┌─┐
                  q_0: ┤ Rp(β) ├┤ Rz(π) ├┤ Rp(β) ├┤ Rz(π) ├─░─┤M├
                       └───────┘└───────┘└───────┘└───────┘ ░ └╥┘
            measure: 1/════════════════════════════════════════╩═
                                                               0

        The parameter β is scanned to find the value that minimizes the unwanted Z-rotation.
        Note that the analysis class requires this experiment to run with three repetition numbers.

    # section: analysis_ref
        :class:`DragCalAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            import warnings

            warnings.filterwarnings(
                "ignore",
                message=".*Due to the deprecation of Qiskit Pulse.*",
                category=DeprecationWarning,
            )

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-.25e9, 1e9, 0.8e9, 1e4, noise=False, seed=101)

        .. jupyter-execute::

            import numpy as np
            from qiskit import pulse
            from qiskit.circuit import Parameter
            from qiskit_experiments.library import RoughDrag

            with pulse.build() as build_sched:
                pulse.play(pulse.Drag(160, 0.50, 40, Parameter("beta")), pulse.DriveChannel(0))

            exp = RoughDrag(physical_qubits=(0,),
                            schedule=build_sched,
                            betas = np.linspace(-4, 4, 51),
                            backend=backend,)
            exp.set_experiment_options(reps=[3, 5, 7])

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)

    # section: reference
        .. ref_arxiv:: 1 1011.1949
        .. ref_arxiv:: 2 0901.0534
        .. ref_arxiv:: 3 1509.05470

    # section: manual
        :ref:`DRAG Calibration`

    """

    @classmethod
    def _default_experiment_options(cls) -> Options:
        r"""Default values for the rough drag experiment.

        Experiment Options:
            schedule (ScheduleBlock): The schedule of the rotation.
            reps (List[int]): The number of times the Rp - Rm gate sequence is repeated in
                each series. Note that this list must always have a length of three as
                otherwise the analysis class will not run.
            betas (Iterable): the values of the DRAG parameter to scan.
        """
        options = super()._default_experiment_options()
        options.schedule = None
        options.reps = [1, 3, 5]
        options.betas = np.linspace(-5, 5, 51)

        return options

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments involving pulse "
            "gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(
        self,
        physical_qubits: Sequence[int],
        schedule: ScheduleBlock,
        betas: Optional[Iterable[float]] = None,
        backend: Optional[Backend] = None,
    ):
        """Initialize a Drag experiment in the given qubit.

        Args:
            physical_qubits: Sequence containing the qubit for which to run the
                Drag calibration.
            schedule: The schedule to run. This schedule should have one free parameter
                corresponding to a DRAG parameter.
            betas: The values of the DRAG parameter to scan. If None is given the default range
                :code:`linspace(-5, 5, 51)` is used.
            backend: Optional, the backend to run the experiment on.

        Raises:
            QiskitError: If the schedule does not have a free parameter.
        """

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="deprecation of Qiskit Pulse",
                module="qiskit_experiments",
                category=DeprecationWarning,
            )
            analysis = DragCalAnalysis()
        super().__init__(physical_qubits, analysis=analysis, backend=backend)

        if betas is not None:
            self.set_experiment_options(betas=betas)

        if len(schedule.parameters) != 1:
            raise QiskitError(
                f"Schedule {schedule} for {self.__class__.__name__} experiment must have "
                f"exactly one free parameter, found {schedule.parameters} parameters."
            )

        self.set_experiment_options(schedule=schedule)

    def _pre_circuit(self) -> QuantumCircuit:
        """A circuit with operations to perform before the Drag."""
        return QuantumCircuit(1)

    def circuits(self) -> List[QuantumCircuit]:
        """Create the circuits for the Drag calibration.

        Returns:
            circuits: The circuits that will run the Drag calibration.

        Raises:
            QiskitError: If the number of different repetition series is not three.
        """
        schedule = self.experiment_options.schedule

        beta = next(iter(schedule.parameters))

        # Note: If the pulse has a reserved name, e.g. x, which does not have parameters
        # then we cannot directly call the gate x and attach a schedule to it. Doing so
        # would results in QObj errors.
        drag_gate = Gate(name="Drag(" + schedule.name + ")", num_qubits=1, params=[beta])

        circuits = []
        for rep in self.experiment_options.reps:
            circuit = self._pre_circuit()
            for _ in range(rep):
                circuit.append(drag_gate, (0,))
                circuit.rz(np.pi, 0)
                circuit.append(drag_gate, (0,))
                circuit.rz(np.pi, 0)

            circuit.measure_active()

            circuit.add_calibration(
                "Drag(" + schedule.name + ")", self.physical_qubits, schedule, params=[beta]
            )

            for beta_val in self.experiment_options.betas:
                beta_val = float(np.round(beta_val, decimals=6))

                assigned_circuit = circuit.assign_parameters({beta: beta_val}, inplace=False)

                assigned_circuit.metadata = {
                    "xval": beta_val,
                    "nrep": rep,
                }

                circuits.append(assigned_circuit)

        return circuits

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
=========================================================
Curve Analysis (:mod:`qiskit_experiments.curve_analysis`)
=========================================================

.. currentmodule:: qiskit_experiments.curve_analysis

Curve analysis provides the analysis base class for a variety of experiments with
a single experimental parameter sweep. This analysis subclasses can override
several class attributes to customize the behavior from data processing to post-processing,
including providing systematic initial guess for parameters tailored to the experiment.


Base Classes
============

.. autosummary::
    :toctree: ../stubs/

    BaseCurveAnalysis
    CurveAnalysis
    CompositeCurveAnalysis

Data Classes
============

.. autosummary::
    :toctree: ../stubs/

    ScatterTable
    CurveFitResult
    ParameterRepr
    FitOptions

Standard Analysis Library
=========================

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/analysis.rst

    BlochTrajectoryAnalysis
    DecayAnalysis
    DampedOscillationAnalysis
    OscillationAnalysis
    ResonanceAnalysis
    GaussianAnalysis
    ErrorAmplificationAnalysis

Fit Functions
=============
.. autosummary::
    :toctree: ../stubs/

    fit_function.cos
    fit_function.cos_decay
    fit_function.exponential_decay
    fit_function.gaussian
    fit_function.sqrt_lorentzian
    fit_function.sin
    fit_function.sin_decay

Initial Guess Estimators
========================
.. autosummary::
    :toctree: ../stubs/

    guess.constant_sinusoidal_offset
    guess.constant_spectral_offset
    guess.exp_decay
    guess.rb_decay
    guess.full_width_half_max
    guess.frequency
    guess.max_height
    guess.min_height
    guess.oscillation_exp_decay

Utilities
=========
.. autosummary::
    :toctree: ../stubs/

    utils.is_error_not_significant
    utils.analysis_result_to_repr
    utils.convert_lmfit_result
    utils.eval_with_uncertainties
    utils.filter_data
    utils.mean_xy_data
    utils.multi_mean_xy_data
    utils.data_sort
    utils.level2_probability
    utils.probability

"""
from .base_curve_analysis import BaseCurveAnalysis
from .curve_analysis import CurveAnalysis
from .composite_curve_analysis import CompositeCurveAnalysis
from .scatter_table import ScatterTable
from .curve_data import (
    CurveFitResult,
    FitOptions,
    ParameterRepr,
)
from .curve_fit import (
    process_curve_data,
    process_multi_curve_data,
)
from . import guess
from . import fit_function
from . import utils

# standard analysis
from .standard_analysis import (
    DecayAnalysis,
    DampedOscillationAnalysis,
    OscillationAnalysis,
    ResonanceAnalysis,
    GaussianAnalysis,
    ErrorAmplificationAnalysis,
    BlochTrajectoryAnalysis,
)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Cross resonance Hamiltonian tomography.
"""

import warnings
from typing import List, Tuple, Sequence, Optional, Type

import numpy as np
from qiskit import pulse, circuit, QuantumCircuit
from qiskit.circuit.parameterexpression import ParameterValueType
from qiskit.exceptions import QiskitError
from qiskit.providers import Backend
from qiskit.utils.deprecation import deprecate_func
from qiskit_experiments.framework import (
    BaseExperiment,
    BackendTiming,
    Options,
)
from qiskit_experiments.library.characterization.analysis import CrossResonanceHamiltonianAnalysis


class CrossResonanceHamiltonian(BaseExperiment):
    r"""Cross resonance Hamiltonian tomography experiment.

    # section: overview

        This experiment assumes the two qubit Hamiltonian in the form

        .. math::

            H = \frac{I \otimes A}{2} + \frac{Z \otimes B}{2}

        where :math:`A` and :math:`B` are linear combinations of
        the Pauli operators :math:`\in {X, Y, Z}`.
        The coefficient of each Pauli term in the Hamiltonian
        can be estimated with this experiment.

        This experiment is performed by stretching the pulse duration of a cross resonance pulse
        and measuring the target qubit by projecting onto the x, y, and z bases.
        The control qubit state dependent (controlled-) Rabi oscillation on the
        target qubit is observed by repeating the experiment with the control qubit
        both in the ground and excited states. The fit for the oscillations in the
        three bases with the two control qubit preparations tomographically
        reconstructs the Hamiltonian in the form shown above.
        See Ref. [1] for more details.

        More specifically, the following circuits are executed in this experiment.

        .. parsed-literal::

            (X measurement)

                 ┌───┐┌────────────────────┐
            q_0: ┤ P ├┤0                   ├────────────────────
                 └───┘│  cr_tone(duration) │┌─────────┐┌────┐┌─┐
            q_1: ─────┤1                   ├┤ Rz(π/2) ├┤ √X ├┤M├
                      └────────────────────┘└─────────┘└────┘└╥┘
            c: 1/═════════════════════════════════════════════╩═
                                                              0

            (Y measurement)

                 ┌───┐┌────────────────────┐
            q_0: ┤ P ├┤0                   ├─────────
                 └───┘│  cr_tone(duration) │┌────┐┌─┐
            q_1: ─────┤1                   ├┤ √X ├┤M├
                      └────────────────────┘└────┘└╥┘
            c: 1/══════════════════════════════════╩═
                                                   0

            (Z measurement)

                 ┌───┐┌────────────────────┐
            q_0: ┤ P ├┤0                   ├───
                 └───┘│  cr_tone(duration) │┌─┐
            q_1: ─────┤1                   ├┤M├
                      └────────────────────┘└╥┘
            c: 1/════════════════════════════╩═
                                             0

        The ``P`` gate on the control qubit (``q_0``) indicates the state preparation.
        Since this experiment requires two sets of sub experiments with the control qubit in the
        excited and ground state, ``P`` will become ``X`` gate or just be omitted, respectively.
        Here ``cr_tone`` is implemented by a single cross resonance tone
        driving the control qubit at the frequency of the target qubit.
        The pulse envelope might be a flat-topped Gaussian implemented by the parametric pulse
        :class:`~qiskit.pulse.library.parametric_pulses.GaussianSquare`.

        This experiment scans the total duration of the cross resonance pulse
        including the pulse ramps at both edges. The pulse shape is defined by the
        :class:`~qiskit.pulse.library.parametric_pulses.GaussianSquare`, and
        an effective length of these Gaussian ramps with :math:`\sigma` can be computed by

        .. math::

            \tau_{\rm edges}' = \sqrt{2 \pi} \sigma,

        which is usually shorter than the actual edge duration of

        .. math::

            \tau_{\rm edges} = 2 r \sigma,

        where the :math:`r` is the ratio of the actual edge duration to :math:`\sigma`.
        This effect must be considered in the following curve analysis to estimate
        interaction rates.

    # section: analysis_ref
        :class:`CrossResonanceHamiltonianAnalysis`

    # section: reference
        .. ref_arxiv:: 1 1603.04821

    # section: manual
        .. ref_website:: Qiskit Textbook 6.7,
            https://github.com/Qiskit/textbook/blob/main/notebooks/quantum-hardware-pulses/hamiltonian-tomography.ipynb
    """

    # Number of CR pulses. The flat top duration per pulse is divided by this number.
    num_pulses = 1

    class CRPulseGate(circuit.Gate):
        """A pulse gate of cross resonance. Definition should be provided via calibration."""

        def __init__(self, width: ParameterValueType):
            super().__init__("cr_gate", 2, [width])

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments involving pulse "
            "gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(
        self,
        physical_qubits: Tuple[int, int],
        backend: Optional[Backend] = None,
        cr_gate: Optional[Type[circuit.Gate]] = None,
        durations: Optional[Sequence[int]] = None,
        **kwargs,
    ):
        """Create a new experiment.

        Args:
            physical_qubits: Two-value tuple of qubit indices on which to run tomography.
                The first index stands for the control qubit.
            backend: Optional, the backend to run the experiment on.
            cr_gate: Optional, circuit gate class representing the cross resonance pulse.
                Providing this object allows us to run this experiment with circuit simulator,
                and this object might be used for testing, development of analysis protocol,
                and educational purpose without needing to wait for hardware queueing.
                Note that this instance must provide matrix representation, such as
                unitary gate or Hamiltonian gate, and the class is expected to be instantiated
                with a single parameter ``width`` in units of sec.
            durations: Optional. The total duration of cross resonance pulse(s) including
                rising and falling edges. The minimum number should be larger than the
                total lengths of these ramps. If not provided, then ``num_durations`` evenly
                spaced durations between ``min_durations`` and ``max_durations`` are
                automatically generated from these experiment options. The default numbers
                are chosen to have a good sensitivity for the Hamiltonian coefficient
                of interest at the rate around 1 MHz.
                This argument should be provided in units of sec.
            kwargs: Pulse parameters. See :meth:`experiment_options` for details.

        Raises:
            QiskitError: When ``qubits`` length is not 2.
        """
        if len(physical_qubits) != 2:
            raise QiskitError(
                "Length of qubits is not 2. Please provide index for control and target qubit."
            )

        self._gate_cls = cr_gate or self.CRPulseGate
        self._backend_timing = None

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="deprecation of Qiskit Pulse",
                module="qiskit_experiments",
                category=DeprecationWarning,
            )
            analysis = CrossResonanceHamiltonianAnalysis()
        super().__init__(physical_qubits, analysis=analysis, backend=backend)
        self.set_experiment_options(durations=durations, **kwargs)

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            durations (np.ndarray): The total duration of the cross resonance pulse(s) to scan,
                in units of sec. Values should be longer than pulse ramps.
            min_durations (int): The minimum default pulse duration in samples.
            max_durations (int): The maximum default pulse duration in samples.
            num_durations (int): The number of measured durations. The experiment automatically
                creates durations of linear increment along with ``min_durations`` and
                ``max_durations`` when user doesn't explicitly provide ``durations``.
            amp (complex): Amplitude of the cross resonance tone.
            amp_t (complex): Amplitude of the cancellation or rotary drive on target qubit.
            sigma (float): Sigma of Gaussian rise and fall edges, in units of dt.
            risefall (float): Ratio of edge durations to sigma.
        """
        options = super()._default_experiment_options()
        options.durations = None
        options.min_durations = 60e-9
        options.max_durations = 1200e-9
        options.num_durations = 48
        options.amp = 0.5
        options.amp_t = 0.0
        options.sigma = 64
        options.risefall = 2

        return options

    def _set_backend(self, backend: Backend):
        """Set the backend for the experiment with timing analysis."""
        super()._set_backend(backend)
        self._backend_timing = BackendTiming(backend)

    def _get_dt(self) -> float:
        """A helper function to get finite dt.

        Returns:
            Backend dt value.
        """
        if not self._backend or self._backend_timing.dt is None:
            # When backend timing is not initialized or backend doesn't report dt.
            return 1.0
        return self._backend_timing.dt

    def _get_width(self, duration: ParameterValueType) -> ParameterValueType:
        """A helper function to get flat top width.

        Args:
            duration: Cross resonance pulse duration in units of sec.

        Returns:
            A flat top widths of cross resonance pulse in units of sec.
        """
        sigma_sec = self.experiment_options.sigma * self._get_dt()

        return duration - 2 * sigma_sec * self.experiment_options.risefall

    def _get_durations(self) -> np.ndarray:
        """Return cross resonance pulse durations in units of sec."""
        opt = self.experiment_options

        if opt.durations is None:
            return np.linspace(opt.min_durations, opt.max_durations, opt.num_durations)

        return np.asarray(opt.durations, dtype=float)

    def _build_cr_circuit(self, pulse_gate: circuit.Gate) -> QuantumCircuit:
        """Single tone cross resonance.

        Args:
            pulse_gate: A pulse gate to represent a single cross resonance pulse.

        Returns:
            A circuit definition for the cross resonance pulse to measure.
        """
        cr_circuit = QuantumCircuit(2)
        cr_circuit.append(pulse_gate, [0, 1])

        return cr_circuit

    def _build_default_schedule(self) -> pulse.ScheduleBlock:
        """GaussianSquared cross resonance pulse.

        Returns:
            A schedule definition for the cross resonance pulse to measure.
        """
        opt = self.experiment_options
        duration = circuit.Parameter("duration")

        cr_drive = self._backend_data.control_channel(self.physical_qubits)[0]
        c_drive = self._backend_data.drive_channel(self.physical_qubits[0])
        t_drive = self._backend_data.drive_channel(self.physical_qubits[1])

        with pulse.build(default_alignment="left", name="cr") as cross_resonance:
            # add cross resonance tone
            pulse.play(
                pulse.GaussianSquare(
                    duration=duration,
                    amp=opt.amp,
                    sigma=opt.sigma,
                    risefall_sigma_ratio=opt.risefall,
                ),
                cr_drive,
            )
            # add cancellation tone
            if not np.isclose(opt.amp_t, 0.0):
                pulse.play(
                    pulse.GaussianSquare(
                        duration=duration,
                        amp=opt.amp_t,
                        sigma=opt.sigma,
                        risefall_sigma_ratio=opt.risefall,
                    ),
                    t_drive,
                )
            else:
                pulse.delay(duration, t_drive)

            # place holder for empty drive channels. this is necessary due to known pulse gate bug.
            pulse.delay(duration, c_drive)

        return cross_resonance

    def circuits(self) -> List[QuantumCircuit]:
        """Return a list of experiment circuits.

        Returns:
            A list of :class:`QuantumCircuit`.

        Raises:
            QiskitError: When the backend is not set and cr gate is ``CRPulseGate`` type.
        """
        if self._gate_cls is self.CRPulseGate:
            if not self.backend:
                # Backend is not set, but trying to provide CR gate as a pulse gate.
                raise QiskitError(
                    "This experiment requires to have backend set to convert durations into samples "
                    "with backend reported dt value and also it requires the channel mapping from "
                    "the backend to build cross resonance pulse schedule. "
                    "Please provide valid backend object supporting 2Q pulse gate."
                )
            return self._pulse_gate_circuits()
        return self._unitary_circuits()

    def _pulse_gate_circuits(self):
        """Protocol to create circuits with pulse gate.

        Pulse gate has backend timing constraints and duration should be in units of dt.
        This method calls :meth:`_build_default_schedule` to generate actual schedule.
        We assume backend has been set in this method call.
        """
        schedule = self._build_default_schedule()

        # Assume this parameter is in units of dt, because this controls pulse samples.
        param_duration = next(iter(schedule.get_parameters("duration")))

        # Gate duration will be shown in sec, which is more intuitive.
        cr_gate = self._gate_cls(width=self._get_width(self._backend_timing.dt * param_duration))

        # Create parameterized circuits with calibration.
        tmp_circs = []
        for control_state in (0, 1):
            for meas_basis in ("x", "y", "z"):
                tmp_qc = QuantumCircuit(2, 1)
                if control_state:
                    tmp_qc.x(0)
                tmp_qc.compose(
                    other=self._build_cr_circuit(cr_gate),
                    qubits=[0, 1],
                    inplace=True,
                )
                if meas_basis == "x":
                    tmp_qc.rz(np.pi / 2, 1)
                if meas_basis in ("x", "y"):
                    tmp_qc.sx(1)
                tmp_qc.measure(1, 0)
                tmp_qc.metadata = {
                    "control_state": control_state,
                    "meas_basis": meas_basis,
                }
                tmp_qc.add_calibration(cr_gate, self.physical_qubits, schedule)
                tmp_circs.append(tmp_qc)

        circs = []
        for duration in self._get_durations():
            # Need to round pulse to satisfy hardware timing constraints.
            # Convert into samples for assignment and validation.
            valid_duration_dt = self._backend_timing.round_pulse(time=duration)

            # Convert into sec to pass xval to analysis.
            # Analysis expects xval of flat top widths in units of sec.
            flat_top_width_sec = self._get_width(self._backend_timing.dt * valid_duration_dt)
            if flat_top_width_sec < 0:
                raise ValueError(
                    f"Input duration={duration} is less than pulse ramps lengths, resulting in "
                    f"a negative flat top length of {flat_top_width_sec} sec. "
                    f"This cross resonance schedule is invalid."
                )

            for circ in tmp_circs:
                # Assign duration in dt to create pulse schedule.
                assigned_circ = circ.assign_parameters(
                    {param_duration: valid_duration_dt},
                    inplace=False,
                )
                assigned_circ.metadata["xval"] = self.num_pulses * flat_top_width_sec
                circs.append(assigned_circ)

        return circs

    def _unitary_circuits(self):
        """Protocol to create circuits with unitary gate.

        Unitary gate has no timing constraints and accepts duration in sec.
        Basically, this method doesn't require backend apart from conversion of
        sigma in samples into sec.
        """
        # Assume this parameter is in units of sec.
        param_duration = circuit.Parameter("duration")

        # Gate duration will be shown in sec, which is more intuitive.
        cr_gate = self._gate_cls(width=self._get_width(param_duration))

        # Create parameterized circuits without calibration.
        tmp_circs = []
        for control_state in (0, 1):
            for meas_basis in ("x", "y", "z"):
                tmp_qc = QuantumCircuit(2, 1)
                if control_state:
                    tmp_qc.x(0)
                tmp_qc.compose(
                    other=self._build_cr_circuit(cr_gate),
                    qubits=[0, 1],
                    inplace=True,
                )
                if meas_basis == "x":
                    tmp_qc.rz(np.pi / 2, 1)
                if meas_basis in ("x", "y"):
                    tmp_qc.sx(1)
                tmp_qc.measure(1, 0)
                tmp_qc.metadata = {
                    "control_state": control_state,
                    "meas_basis": meas_basis,
                }
                tmp_circs.append(tmp_qc)

        circs = []
        for duration in self._get_durations():
            flat_top_width_sec = self._get_width(duration)
            if flat_top_width_sec < 0:
                raise ValueError(
                    f"Input duration={duration} is less than pulse ramps lengths, resulting in "
                    f"a negative flat top length of {flat_top_width_sec} sec. "
                    f"This cross resonance schedule is invalid."
                )

            for circ in tmp_circs:
                # Assign duration in sec since this is unitary gate.
                assigned_circ = circ.assign_parameters(
                    {param_duration: duration},
                    inplace=False,
                )
                assigned_circ.metadata["xval"] = self.num_pulses * flat_top_width_sec
                circs.append(assigned_circ)

        return circs

    def _finalize(self):
        """Set analysis option for initial guess that depends on experiment option values."""
        edge_duration = np.sqrt(2 * np.pi) * self.experiment_options.sigma * self.num_pulses

        for analysis in self.analysis.analyses():
            init_guess = analysis.options.p0.copy()
            if "t_off" in init_guess:
                continue
            init_guess["t_off"] = self._get_dt() * edge_duration
            analysis.set_options(p0=init_guess)

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata


class EchoedCrossResonanceHamiltonian(CrossResonanceHamiltonian):
    r"""Echoed cross resonance Hamiltonian tomography experiment.

    # section: overview

        This is a variant of :class:`CrossResonanceHamiltonian`
        for which the experiment framework is identical but the
        cross resonance operation is realized as an echoed sequence
        to remove unwanted single qubit rotations. The cross resonance
        circuit looks like:

        .. parsed-literal::

                 ┌────────────────────┐  ┌───┐  ┌────────────────────┐
            q_0: ┤0                   ├──┤ X ├──┤0                   ├──────────
                 │  cr_tone(duration) │┌─┴───┴─┐│  cr_tone(duration) │┌────────┐
            q_1: ┤1                   ├┤ Rz(π) ├┤1                   ├┤ Rz(-π) ├
                 └────────────────────┘└───────┘└────────────────────┘└────────┘

        Here two ``cr_tone`` are applied, where the latter one is with the
        control qubit state flipped and with a phase flip of the target qubit frame.
        This operation is equivalent to applying the ``cr_tone`` with a negative amplitude.
        The Hamiltonian for this decomposition has no IX and ZI interactions,
        and also a reduced IY interaction to some extent (not completely eliminated) [1].
        Note that the CR Hamiltonian tomography experiment cannot detect the ZI term.
        However, it is sensitive to the IX and IY terms.

    # section: reference
        .. ref_arxiv:: 1 2007.02925

    """

    num_pulses = 2

    def _build_cr_circuit(self, pulse_gate: circuit.Gate) -> QuantumCircuit:
        """Single tone cross resonance.

        Args:
            pulse_gate: A pulse gate to represent a single cross resonance pulse.

        Returns:
            A circuit definition for the cross resonance pulse to measure.
        """
        cr_circuit = QuantumCircuit(2)
        cr_circuit.append(pulse_gate, [0, 1])
        cr_circuit.x(0)
        cr_circuit.rz(np.pi, 1)
        cr_circuit.append(pulse_gate, [0, 1])
        cr_circuit.rz(-np.pi, 1)

        return cr_circuit



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
===============================================================================================
Randomized Benchmarking Experiments (:mod:`qiskit_experiments.library.randomized_benchmarking`)
===============================================================================================

.. currentmodule:: qiskit_experiments.library.randomized_benchmarking

Experiments
===========
.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/experiment.rst

    StandardRB
    InterleavedRB
    LayerFidelity


Analysis
========

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/analysis.rst

    RBAnalysis
    InterleavedRBAnalysis
    LayerFidelityAnalysis

Synthesis
=========

.. autosummary::
    :toctree: ../stubs/

    RBDefaultCliffordSynthesis

Utilities
=========

.. autosummary::
    :toctree: ../stubs/

    RBUtils
    CliffordUtils
"""
from .standard_rb import StandardRB
from .interleaved_rb_experiment import InterleavedRB
from .rb_analysis import RBAnalysis
from .interleaved_rb_analysis import InterleavedRBAnalysis
from .clifford_utils import CliffordUtils
from .rb_utils import RBUtils
from .clifford_synthesis import RBDefaultCliffordSynthesis
from .layer_fidelity import LayerFidelity
from .layer_fidelity_analysis import LayerFidelityAnalysis



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Half angle calibration."""

from typing import Dict, Optional, Sequence

from qiskit.circuit import QuantumCircuit
from qiskit.providers.backend import Backend

from qiskit_experiments.framework import ExperimentData
from qiskit_experiments.exceptions import CalibrationError
from qiskit_experiments.calibration_management import (
    BaseCalibrationExperiment,
    Calibrations,
)
from qiskit_experiments.library.characterization import HalfAngle
from qiskit_experiments.calibration_management.update_library import BaseUpdater


class HalfAngleCal(BaseCalibrationExperiment, HalfAngle):
    """Calibration version of the :class:`.HalfAngle` experiment."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        backend: Optional[Backend] = None,
        schedule_name: str = "sx",
        cal_parameter_name: Optional[str] = "angle",
        auto_update: bool = True,
    ):
        """The experiment to update angle of half-pi rotation gates.

        Args:
            physical_qubits: Sequence containing the qubit for which to run the
                half-angle calibration.
            calibrations: The calibrations instance with the schedules.
            backend: Optional, the backend to run the experiment on.
            schedule_name: The name of the schedule to calibrate which defaults to sx.
            cal_parameter_name: The name of the parameter in the schedule to update. This will
                default to 'angle' in accordance with the naming convention of the
                :class:`~qiskit.pulse.ScalableSymbolicPulse` class.
            auto_update:  Whether or not to automatically update the calibrations. By
                default this variable is set to True.

        Raises:
            CalibrationError: if cal_parameter_name is set to ``amp``, to reflect the
                transition from calibrating complex amplitude to calibrating the phase.
            CalibrationError: if the default cal_parameter_name is used, and it is not
                a valid parameter of the calibrated schedule.
        """
        if cal_parameter_name == "amp":
            raise CalibrationError(
                "The Half-Angle calibration experiment was changed from calibrating"
                " the pulse's complex amplitude, to calibrating the angle parameter "
                "in the real (amp,angle) representation. Setting cal_parameter_name to "
                "'amp' thus indicates that you are probably using the experiment in "
                "an inconsistent way. If your pulse does in fact use a complex amplitude,"
                "you need to convert it to (amp,angle) representation, preferably using"
                "the ScalableSymbolicPulse class. Note that all library pulses now use "
                "this representation."
            )
        # If the default cal_parameter_name is used, validate that it is in fact a parameter
        if cal_parameter_name == "angle":
            try:
                calibrations.calibration_parameter("angle", schedule_name=schedule_name)
            except CalibrationError as err:
                raise CalibrationError(
                    "The Half-Angle calibration experiment was changed from calibrating"
                    " the pulse's complex amplitude, to calibrating the angle parameter "
                    "in the real (amp,angle) representation. The default cal_parameter_name "
                    "was thus changed to angle, which is not a valid parameter of the "
                    "calibrated schedule. It is likely that you are trying to calibrate "
                    "a schedule which is defined by a complex amplitude. To use the "
                    "Half-Angle experiment you need to convert the pulses in the schedule "
                    "to (amp,angle) representation (preferably, using the "
                    "ScalableSymbolicPulse class), and have a parameter associated with "
                    "the angle. Note that all library pulses now use this representation."
                ) from err

        super().__init__(
            calibrations,
            physical_qubits,
            backend=backend,
            schedule_name=schedule_name,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )

    def _metadata(self) -> Dict[str, any]:
        """Add metadata to the experiment data making it more self contained.

        The following keys are added to the experiment's metadata:
            cal_param_value: The value of the pulse amplitude. This value together with
                the fit result will be used to find the new value of the pulse amplitude.
            cal_param_name: The name of the parameter in the calibrations.
            cal_schedule: The name of the schedule in the calibrations.
            cal_group: The calibration group to which the parameter belongs.
        """
        metadata = super()._metadata()
        metadata["cal_param_value"] = self._cals.get_parameter_value(
            self._param_name,
            self._physical_qubits,
            self._sched_name,
            group=self.experiment_options.group,
        )

        return metadata

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Adds the calibrations to the transpiled circuits."""
        for gate in ["y", "sx"]:
            schedule = self._cals.get_schedule(gate, self.physical_qubits)
            circuit.add_calibration(gate, self.physical_qubits, schedule)

    def update_calibrations(self, experiment_data: ExperimentData):
        r"""Update the value of the parameter in the calibrations.

        The parameter that is updated is the phase of the sx pulse. This phase is contained
        in the complex amplitude of the pulse. The update rule for the half angle calibration is
        therefore:

        .. math::

            A \to A \cdot e^{-i{\rm d}\theta_\text{hac}/2}

        where :math:`A` is the complex amplitude of the sx pulse which has an angle which might be
        different from the angle of the x pulse due to the non-linearity in the mixer's skew. The
        angle :math:`{\rm d}\theta_\text{hac}` is the angle deviation measured through the error
        amplifying pulse sequence.

        Args:
            experiment_data: The experiment data from which to extract the measured over/under
                rotation used to adjust the amplitude.
        """

        result_index = self.experiment_options.result_index
        group = experiment_data.metadata["cal_group"]
        prev_angle = experiment_data.metadata["cal_param_value"]

        d_theta = BaseUpdater.get_value(experiment_data, "d_hac", result_index)
        new_angle = prev_angle - (d_theta / 2)

        BaseUpdater.add_parameter_value(
            self._cals,
            experiment_data,
            new_angle,
            self._param_name,
            self._sched_name,
            group,
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Half angle characterization."""

from typing import List, Optional, Sequence
import numpy as np

from qiskit import QuantumCircuit
from qiskit.providers import Backend

from qiskit_experiments.framework import BaseExperiment, Options
from qiskit_experiments.curve_analysis.standard_analysis import ErrorAmplificationAnalysis
from qiskit_experiments.curve_analysis import ParameterRepr


class HalfAngle(BaseExperiment):
    r"""An experiment class to measure the amount by which sx and x are not parallel.

    # section: overview

        This experiment runs circuits that repeat blocks of :code:`sx - sx - y` gates
        inserted in a Ramsey type experiment, i.e. the full gate sequence is thus
        :code:`Ry(π/2) - [sx - sx - y] ^ n - sx` where :code:`n` is varied.

        .. parsed-literal::

                    ┌─────────┐┌────┐┌────┐┌───┐   ┌────┐┌────┐┌───┐┌────┐ ░ ┌─┐
               q_0: ┤ Ry(π/2) ├┤ sx ├┤ sx ├┤ y ├...┤ sx ├┤ sx ├┤ y ├┤ sx ├─░─┤M├
                    └─────────┘└────┘└────┘└───┘   └────┘└────┘└───┘└────┘ ░ └╥┘
            meas: 1/════════════════════════════...═══════════════════════════╩═
                                                                              0

        This sequence measures angle errors where the axis of the :code:`sx` and :code:`x`
        rotation are not parallel. A similar experiment is described in Ref.~[1] where the
        gate sequence :code:`x - y` is repeated to amplify errors caused by non-orthogonal
        :code:`x` and :code:`y` rotation axes.

        One cause of such errors is non-linearity in the microwave mixer used
        to produce the pulses for the ``x`` and ``sx`` gates. Typically, these
        gates are calibrated to have the same duration and so have different
        pulse amplitudes. Non-linearities in the mixer's skew can cause the
        angle to differ for these different pulse amplitudes.

        The way the experiment works is that the initial ``Ry(π/2)`` puts the
        qubit close to the :math:`+X` state, with a deviation :math:`δθ`, due
        to the misalignment between ``sx`` and ``x`` (``Ry(π/2)`` is
        implemented with ``sx`` as described below). The first ``sx - sx`` do
        nothing as they should be rotations about the axis the qubit is
        pointing along. The first ``y`` then mirrors the qubit about the
        :math:`y` axis in the :math:`xy` plane of the Bloch sphere, so the
        :math:`δθ` deviation from :math:`+X` becomes a :math:`-δθ` from
        :math:`-X`. The next ``sx - sx`` sequence rotates about the axis that
        is :math:`+δθ` rotated in the :math:`xy` plane from :math:`+X`, which
        takes the deviation from :math:`-X` from :math:`-δθ` to :math:`+3 δθ`.
        Then the next ``y`` mirrors this across the :math:`y` axis, taking the
        state to :math:`-3 δθ` from :math:`+X`. This pattern continues with
        each iteration, with the angular deviation in units of :math:`δθ`
        following the sequence 1, 3, 5, 7, 9, etc. from :math:`+X` and
        :math:`-X`. The final ``sx`` rotation serves mainly to rotate these
        deviations from :math:`+X` and :math:`-X` in the :math:`xy` plane into
        deviations out of the :math:`xy` plane, so that they appear as a signal
        in the :math:`Z` basis.  Because ``sx`` has a :math:`δθ` deviation from
        ``x``, the final ``sx`` adds an extra :math:`δθ` to the deviations, so
        the pattern ends up as 2, 4, 6, 8, etc., meaning that each iteration
        adds :math:`2 δθ` to the deviation from the equator of the Bloch sphere
        (with the sign alternating due to the ``y`` gates, so the deviations
        are really -2, 4, -6, 8, etc.).

        For the implementation of the circuits, the experiment uses ``Rz(π/2) -
        sx - Rz(-π/2)`` to implement the ``Ry(π/2)`` and ``Rz(π/2) - x -
        Rz(-π/2)`` to implement the ``y``. So the experiment makes use of only
        ``sx``, ``x``, ``Rz(π/2)``, and ``Rz(-π/2)`` gates. For the
        experiment's analysis to be valid, it is important that the ``sx`` and
        ``x`` gates are not replaced (such as by a transpiler pass that
        replaces ``x`` with ``sx - sx``), as it is the angle between them which
        is being inferred. It is assumed that the angle between ``x`` and
        ``Rz`` is exactly :math:`π/2`.

    # section: analysis_ref
        :class:`.ErrorAmplificationAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-.25e9, 1e9, 0.8e9, 1e4, noise=False, seed=199)

        .. jupyter-execute::

            from qiskit_experiments.library.characterization import HalfAngle

            exp = HalfAngle((0,), backend=backend)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)

    # section: reference
        .. ref_arxiv:: 1 1504.06597
    """

    @classmethod
    def _default_experiment_options(cls) -> Options:
        r"""Default values for the half angle experiment.

        Experiment Options:
            repetitions (List[int]): A list of the number of times that the gate
                sequence :code:`[sx sx y]` is repeated.
        """
        options = super()._default_experiment_options()
        options.repetitions = list(range(15))
        return options

    def __init__(self, physical_qubits: Sequence[int], backend: Optional[Backend] = None):
        """Setup a half angle experiment on the given qubit.

        Args:
            physical_qubits: List containing the qubits on which to run the
                fine amplitude calibration experiment.
            backend: Optional, the backend to run the experiment on.
        """
        analysis = ErrorAmplificationAnalysis()

        default_bounds = analysis.options.bounds
        default_bounds.update({"d_theta": (-np.pi / 2, np.pi / 2)})

        analysis.set_options(
            fixed_parameters={
                "angle_per_gate": np.pi,
                "phase_offset": -np.pi / 2,
                "amp": 1.0,
            },
            result_parameters=[ParameterRepr("d_theta", "d_hac", "rad")],
            normalization=True,
            bounds=default_bounds,
        )

        super().__init__(physical_qubits, analysis=analysis, backend=backend)

    @staticmethod
    def _pre_circuit() -> QuantumCircuit:
        """Return the preparation circuit for the experiment."""
        return QuantumCircuit(1)

    def circuits(self) -> List[QuantumCircuit]:
        """Create the circuits for the half angle calibration experiment."""

        circuits = []

        for repetition in self.experiment_options.repetitions:
            circuit = self._pre_circuit()

            # First ry gate
            circuit.rz(np.pi / 2, 0)
            circuit.sx(0)
            circuit.rz(-np.pi / 2, 0)

            # Error amplifying sequence
            for _ in range(repetition):
                circuit.sx(0)
                circuit.sx(0)
                circuit.rz(np.pi / 2, 0)
                circuit.x(0)
                circuit.rz(-np.pi / 2, 0)

            circuit.sx(0)
            circuit.measure_all()

            circuit.metadata = {"xval": repetition}

            circuits.append(circuit)

        return circuits

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata
